/*
 * botactivity.cpp
 *
 *  Created on: 17 Sep 2015
 *      Author: irina
 */


#include <psconfig.h>
#include "globals.h"
#include <csgeom/math3d.h>

#include "net/connection.h"
#include "util/log.h"
#include "util/strutil.h"
#include "gui/inventorywindow.h"


#include "botactivity.h"

MovementStrategy::MovementStrategy()
{
	size = DEFAULT_SIZE;
	strategy = DEFAULT_STRATEGY;
	startPos = MIDDLE_MAP_POS;
}

void MovementStrategy::Move(int movesDone)
{
	switch (strategy)
	{
	case EXPLORE_RADIAL:
	{
		if (movesDone >0 && movesDone % size ==0)
			if ((movesDone/size) % 2 == 1)
				worldHandler::TurnToPoint(startPos.x, startPos.z);
			else worldHandler::Turn(ANGLE_RADIAL_STRATEGY);
		worldHandler::StepForward();
		break;
	}
	case EXPLORE_GRID:
	{
		if (movesDone>0 && (movesDone+1) % size ==0)
			worldHandler::Move(MVM_FORWARD);
		else if (((movesDone+1) / size) % 2 == 0)
			worldHandler::Move(MVM_STRAFE_RIGHT);
		else worldHandler::Move(MVM_STRAFE_LEFT);

		break;
	}
	case EXPLORE_LINE:
	{
		if (((movesDone+1) / size) % 2 ==0)
			worldHandler::StepForward();
		else worldHandler::StepBackward();

		break;
	}
	default: worldHandler::StepForward();
	}
}


void MovementStrategy::UnMove(int movesDone)
{
	switch (strategy)
	{
	case EXPLORE_GRID:
	{
		if (movesDone>0 && (movesDone+1) % size ==0)
			worldHandler::Move(MVM_BACKWARD);
		else if (((movesDone+1) / size) % 2 == 0)
			worldHandler::Move(MVM_STRAFE_LEFT);
		else worldHandler::Move(MVM_STRAFE_RIGHT);

		break;
	}
	case EXPLORE_LINE:
	{
		if (((movesDone+1) / size) % 2 ==0)
			worldHandler::StepBackward();
		else worldHandler::StepForward();

		break;
	}
	default: worldHandler::StepBackward();
	}
}


ExploreActivity::ExploreActivity()
{
	name = csString("explore");
	textcolour = TEXT_COLOUR;
	exploreInvCache = new psInventoryCache();
	watchingInvForNewResources = false;
	timesDone = -1;
}
ExploreActivity::~ExploreActivity()
{

}
void ExploreActivity::HandleMessage(MsgEntry* message )
{
    if (MSGTYPE_DEAD_RECKONING == message->GetType())
	{
		if (IsOngoing() && shouldBeStandingStill)
		{
			worldHandler::StopAllMovement();
			float yRot = worldHandler::GetCurrentRot();
			psDRMessage drmsg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers() );
			csVector3 serverPos = drmsg.pos;
			psengine->GetCelClient()->GetMainPlayer()->SetPosition(serverPos, drmsg.yrot, drmsg.sector);
			printf("Matching server position.\n");
			OutputMsg(csString("    matching server position"));
			worldHandler::SetRotation(yRot);
			worldHandler::SendDRUpdate();
		}
	}
    else if (MSGTYPE_GUIINVENTORY == message->GetType())
	{
		psGUIInventoryMessage incoming(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());

		// drop inventory list, if its version is older than the current.
		// this may happen due to UDP latency.
		if (exploreInvCache->GetInventoryVersion() >= incoming.version)
		{
			message->Reset();
			return;
		}
    
		// Set new version in any case (LIST or UPDATE_LIST)
		exploreInvCache->SetInventoryVersion(incoming.version);

		// merge the received inventory list into the client's inventory cache
		if (incoming.command == psGUIInventoryMessage::LIST)   // for the whole lot
		{
			exploreInvCache->EmptyInventory();   // empty the inventory first
			exploreInvCache->SetCacheStatus(psCache::VALID);
		}

		int tableCount = 0;

		for ( size_t z = 0; z < incoming.totalItems; z++ )  // cache inventory items
		{
			exploreInvCache->SetInventoryItem(incoming.items[z].slot,
											  incoming.items[z].container,
											  incoming.items[z].name,
											  incoming.items[z].meshName,
											  incoming.items[z].materialName,
											  incoming.items[z].weight,
											  incoming.items[z].size,
											  incoming.items[z].stackcount,
											  incoming.items[z].itemQuality,
											  incoming.items[z].iconImage,
											  incoming.items[z].purifyStatus);

			csString name = incoming.items[z].name;
			if (name.CompareNoCase(CRAFT_TABLE))
				tableCount++;
		}

		if (usingTable && tableCount == initialTableCount && !pickupDone)
		{
				pickupDone = true;
				tableEid = EID(-1001);
				lastKnownTableDist = -1.0f;
				if (seekingTableUrgently)
				{
					worldHandler::SetRotation(prevRotation);
					seekingTableUrgently = false;
				}
				if (glitchPickup)
				{
					glitchPickup = false;
				}
				else
				{
					Finish();
				}
		}


		if (incoming.command == psGUIInventoryMessage::UPDATE_LIST)
		{
			for ( size_t z = 0; z < incoming.totalEmptiedSlots; z++)
			{
				exploreInvCache->EmptyInventoryItem( incoming.items[incoming.totalItems+z].slot,
													incoming.items[incoming.totalItems+z].container);
			}
		}

		if (watchingInvForNewResources)
		{
			watchingInvForNewResources = false;
			csArray<slot> toProcess = csArray<slot>();
			for (size_t i = 0; i < invSnapshot.GetSize(); i++)
			{
				psInventoryCache::CachedItemDescription* cacheSlot = exploreInvCache->GetInventoryItem(i+16);

				if (!invSnapshot[i] && cacheSlot != NULL)
				{
					slot tempSlot = slot(cacheSlot->slot, cacheSlot->name, cacheSlot->stackCount, cacheSlot->itemQuality, true);
					tempSlot.container = cacheSlot->containerID;
					int64_t count = cacheSlot->stackCount;
					csString itemName = cacheSlot->name;
					claimOutput.PutUnique(itemName, count + claimOutput.Get(itemName, 0));
					OutputMsg(csString("    ").Append(count).Append(" x ").Append(itemName).Append(" q").Append(cacheSlot->itemQuality),TEXT_COLOUR_WHITE);
					toProcess.Push(tempSlot);
				}
			}
			// if toProcess is empty, resource moves are done
			if (toProcess.IsEmpty())
			{
				OutputMsg(csString("    no resources found"));
			}
			else
			{
				while (!toProcess.IsEmpty())
				{
					printf("toProcess listing:\n");
					printSlots(toProcess);
					csArray<slot> currentBand = csArray<slot>();
					currentBand.Push(toProcess.Pop());
					if (!toProcess.IsEmpty())
					{
						// check for same-band items that also need processing
						int currentBandNo = currentBand.Get(0).quality / qualityBand;
						csArray<slot> toProcessCopy = csArray<slot>(toProcess);
						for (size_t i = 0; i < toProcessCopy.GetSize(); i++)
						{
							slot cohort = toProcessCopy.Get(i);
							if ((cohort.quality / qualityBand) == currentBandNo && cohort.name.Compare(currentBand.Get(0).name))
							{
								currentBand.Push(cohort);
								toProcess.Delete(cohort);
							}
						}
					}
					printSlots(currentBand);
					csArray<move> moveList = Stack(currentBand);
					for (size_t i = 0; i < moveList.GetSize(); i++)
					{
						move nextMove = moveList.Get(i);
						printf("queueing move from Stack():\n\t");
						printMove(nextMove);
						pendingResourceMoves.PushBack(nextMove);
					}
					
				}
			}
		}
		else if (verifyInventoryMove)
		{
			int targetCount = currentResourceMove.dest.count + currentResourceMove.count;
			int targetSlot = currentResourceMove.dest.slotNo;
			psInventoryCache::CachedItemDescription* cacheSlot = exploreInvCache->GetInventoryItem(targetSlot);
			if (cacheSlot->stackCount == targetCount)
			{
				//OutputMsg(csString("    inventory slot move verified"));
				verifyInventoryMove = false;
				int targetQual = (currentResourceMove.src.quality + currentResourceMove.dest.quality) / 2;
				if (!currentResourceMove.src.isInInventory)
				{
					tableContents[currentResourceMove.src.slotNo].count -= currentResourceMove.count;
				}
				lastVerifyTime = csGetTicks();
				storingResourcesInProgress = false;
			}
			else
			{
				OutputMsg(csString("  expecting inventory slot ").Append(targetSlot).Append(" to have ").Append(targetCount).Append(" items, not ").Append(cacheSlot->stackCount));
			}
		}

	} 
	else if (MSGTYPE_VIEW_CONTAINER == message->GetType())
	{
		if (needTableScan)
		{
			needTableScan = false;
			// clear cache of table slot contents
			for (int i = 0; i < 16; i++)
			{
				tableContents[i].Clear(i);
			}
			psViewContainerDescription mesg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
			if (tableEid.Unbox() == -1001)
			{
				tableEid = EID(mesg.containerID);
			}
			OutputMsg(csString("Scanning table..."));
			if ( mesg.hasContents && mesg.contents.GetSize() > 0)
			{
				tableScanSlots.DeleteAll();
				for (size_t i=0; i < mesg.contents.GetSize(); i++)
				{
					int slotID = mesg.contents[i].slotID;
					tableContents[slotID].name = mesg.contents[i].name;
					tableContents[slotID].count = mesg.contents[i].stackCount;
					tableContents[slotID].container = mesg.containerID;
					tableScanSlots.InsertSorted(slotID);
				}
				psViewItemDescription out(mesg.containerID, tableScanSlots.Get(0));
				out.SendMessage();
			}
			else
			{
				OutputMsg(csString("    { all slots empty }"));
				tableScanInProgress = false;
				DoSendTableTest();
			}
		}
		else if (tableTestSent)
		{
			tableTestSent = false;
			tableTestPassed = true;
		}
		else if (verifyTableSlotMove)
		{
			psViewContainerDescription mesg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
			int targetCount = currentResourceMove.dest.count + currentResourceMove.count;
			int targetSlot = currentResourceMove.dest.slotNo;
			for (size_t i=0; i < mesg.contents.GetSize(); i++)
			{
				int slotID = mesg.contents[i].slotID;
				if (slotID == targetSlot)
				{
					int stackCount = mesg.contents[i].stackCount;
					if (stackCount == targetCount)
					{
						//OutputMsg(csString("    table slot move verified"));
						verifyTableSlotMove = false;
						int targetQual = (currentResourceMove.src.quality + currentResourceMove.dest.quality) / 2;
						tableContents[targetSlot].quality = targetQual;
						tableContents[targetSlot].count = targetCount;
						if (!currentResourceMove.src.isInInventory)
						{
							tableContents[currentResourceMove.src.slotNo].count -= currentResourceMove.count;
						}
						lastVerifyTime = csGetTicks();
						storingResourcesInProgress = false;
					}
					else
					{
						OutputMsg(csString("  expecting table slot ").Append(targetSlot).Append(" to have ").Append(targetCount).Append(" items, not ").Append(stackCount));
					}
				}
			}

		}
	}
	else if (MSGTYPE_VIEW_ITEM == message->GetType())
	{
		psViewItemDescription mesg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
		csString desc(mesg.itemDescription);
		desc = desc.Collapse();
		if (!tableScanSlots.IsEmpty())
		{
			int slotID = tableScanSlots.Get(0);
			tableScanSlots.DeleteIndex(0);
			size_t start = desc.Find("Quality: ") + 9;
			size_t length = desc.Find(" ", start) - start;
			csString qualityStr = desc.Slice(start, length).Trim();
			int quality = atoi(qualityStr.GetData());
			tableContents[slotID].quality = quality;
			OutputMsg(csString("    slot ").Append(slotID).Append(": ").Append(tableContents[slotID].count).Append(" x ").Append(mesg.itemName).Append(" q").Append(quality));
			if (tableScanSlots.IsEmpty())
			{
				tableScanInProgress = false;
				DoSendTableTest();
			}
			else
			{
				psViewItemDescription out(tableEid.Unbox(), tableScanSlots.Get(0));
				out.SendMessage();
			}
		}
		else if (parseClaimKey)
		{
			size_t start = desc.Find("Description: Key to A");
			if (start == (size_t)-1)
			{
				printf("not a key description: %s\n", desc.GetData());
				message->Reset();
				return;
			}
			start += 22;
			size_t end = desc.Find(" ", start);
			csString keySize = desc.Slice(start, end-start);
			start = end + 1;
			end = desc.Find(" Claim at ", end);
			csString keyItem = desc.Slice(start, end-start);
			csString locStr = desc.Slice(end+10, desc.Length());
			WordArray loc = WordArray(locStr);
			float x = loc.GetFloat(0);
			float y = loc.GetFloat(1);
			float z = loc.GetFloat(2);
			parseClaimKey = false;
			markerPos = csVector3(x,y,z);
			printf("saving key pos %f, %f, %f, (%s)\n",x, y, z, markerPos.Description().GetData());

		}
	}
	if (!IsOngoing())
	{
		message->Reset();
		return;
	}

	switch (message->GetType() )
		{
		case MSGTYPE_MODE:
			{
				break;
			}
		case MSGTYPE_CRAFT_INFO:
			{
				psMsgCraftingInfo incoming(message);
				csString itemDescription(incoming.craftInfo);

				claimRcp.ParseRecipe(itemDescription, useMax);
				worldHandler::CloseOpenWindow(csString("Read Book"));
			break;
			}
		case MSGTYPE_SYSTEM:
			{
				psSystemMessage incoming(message);
				csString text = incoming.msgline;

				if (text.StartsWith(LOOT_ON_GROUND))
				{
    				  pickupGroundLoot = true;
				}
				else if (text.StartsWith(CANNOT_BE_PICKED_UP) && !pickupDone)
				{
    				  Stop();
					  OutputMsg(csString("Cannot pickup table, stopping."));
					  exploreFinished = true;
				}
				else if ((actionToDo == ACTION_EXPLORE || actionToDo == ACTION_READRECIPE) && 
						 (text.StartsWith(UNSUCCESSFUL_SINCE_MOVED) || text.StartsWith(NOT_SUCCESSFUL) || text.StartsWith(CANNOT_EXPLORE_TWICE)))
				{
					OutputMsg(csString("    explore failed, moving on"));
					actionToDo = ACTION_MOVE;
				}
				else if ((actionToDo == ACTION_USE_COMBINE || actionToDo == ACTION_EXAMINE_RESULT) && 
						 (text.StartsWith(UNSUCCESSFUL_SINCE_MOVED) || text.StartsWith(NOT_SUCCESSFUL)))
				{
					OutputMsg(csString("    build failed, retrying"));
					rebuild = true;
				}
				else if (text.StartsWith(OVERWEIGHT))
				{
                    if (!usingTable)
                    {
    				  Stop();
					  OutputMsg(csString("Explore stopped because of being overweight."));
					  exploreFinished = true;
                    }
				}
				else if (text.StartsWith(DIEING))
				{
					needToDie = true;
					exploreFinished = true;
				}
				else if (text.StartsWith(MAXXED_SKILL))
				{
					OutputMsg(text.Append(": stopping"), TEXT_COLOUR_WHITE);
					noBuild = true;
					exploreFinished = true;
					actionToDo = ACTION_NONE;
					timesToRepeat = timesDone;
					DoStop();
					Stop();
					break;
				}
				else if (text.StartsWith(EXPLORE_TOOL_WORN))
				{
					noBuild = true;
					exploreFinished = true;
					csString toolName = worldHandler::GetEquippedToolName();
					int quality = worldHandler::GetEquippedToolQuality();
					int betterToolSlot = SelectToolSlot(toolName, quality);
					if (betterToolSlot == -1)
					{
						OutputMsg(toolName.Append(" is worn and there is no better replacement, bot will stop!"));
						actionToDo = ACTION_NONE;
						timesToRepeat = timesDone;
						DoStop();
						Stop();
						break;
					}

					if (toolSlot == NULL)
						toolSlot = worldHandler::GetNextFreeInvSlot();

					if (toolSlot == NULL)
					{
						OutputMsg(toolName.Append(" is worn and there is no space to move it, bot will stop!"));
						actionToDo = ACTION_NONE;
						timesToRepeat = timesDone;
						DoStop();
						Stop();
						break;
					}
					
					OutputMsg(toolName.Append(" is worn. Replacing with better tool from slot ").Append(betterToolSlot));

					worldHandler::MoveItems(CONTAINER_INVENTORY_EQUIPMENT, PSCHARACTER_SLOT_RIGHTHAND, toolSlot->ContainerID(), toolSlot->ID(), 1);
					worldHandler::MoveItems(CONTAINER_INVENTORY_EQUIPMENT, betterToolSlot, CONTAINER_INVENTORY_EQUIPMENT, PSCHARACTER_SLOT_RIGHTHAND, 1);
					actionToDo = ACTION_EXPLORE;
				}
				else if (text.StartsWith(CLAIM_PLACED))
				{
					exploreFinished = true;
					noBuild = false;
					size_t end = text.FindLast('.');
					size_t start = CLAIM_PLACED.Length();
					csString thingName = text.Slice(start, end-start);
					claimName = thingName;

					//and get the item name too if we are at it
					end = claimName.Find(" Exploration Marker");
					start = claimName.Find(" ");
					claimType = claimName.Slice(0, start);
					start++;
					size_t length = end - start;
					itemName = claimName.Slice(start, length);
					OutputMsg(csString("    ").Append(claimType).Append(" claim: ").Append(itemName));
				}
				else if (text.StartsWith(GOT_ITEMS))
				{
					exploreFinished = true;
					size_t end = text.FindLast('!');
					size_t start = GOT_ITEMS.Length();
					csString thingName = text.Slice(start, end-start);

					if (thingName.Find(ITEM_KEY) < thingName.Length())
					{
						keyName = thingName;
						// view the key to verify the claim coords
						parseClaimKey = true;
						psViewItemDescription out(nextFreeSlot->ContainerID(), nextFreeSlot->ID());
						out.SendMessage();
					}
					else
					if (thingName.Find(ITEM_ENUM) < thingName.Length())
					{
						scrollName = thingName;
					}
				}
				else if (text.StartsWith(TOO_FAR_AWAY_TO_PICKUP))
				{

					csVector3 cPos = worldHandler::GetCurrentPos();
					float d = csSquaredDist::PointPoint(cPos, tablePos);

					if (seekingTableUrgently)
					{
						actionToDo = ACTION_MOVE_TOWARD_TABLE;
					}
					else if (lastKnownTableDist < 0)
					{
						actionToDo = ACTION_UNMOVE;
					}
					else
					{
						bool gettingCloser = lastKnownTableDist  > d;
						if (gettingCloser)
						{
							actionToDo = ACTION_UNMOVE;
						}
						else
						{
							seekingTableUrgently = true;
							actionToDo = ACTION_MOVE_TOWARD_TABLE;
							prevRotation = worldHandler::GetCurrentRot();
						}
					}
					lastKnownTableDist = d;
				}
				else if (text.StartsWith(TOO_FAT_TO_PICKUP))
				{
					OutputMsg("Too fat to pickup table. Table may be glitched.");
					DoStop();
					Stop();
				}
				else if (text.StartsWith(TOO_FAR_AWAY_TO_USE) || text.StartsWith(IT_DOESNT_FIT_THERE) || text.StartsWith(CANT_MOVE_ITEMS))
				{
					if (verifyClaimTransfer)
					{
						verifyClaimTransfer = false;
						claimTransferSuccess = false;
					}
				}
				else if (text.StartsWith(YOU_DROPPED))
				{
				}
				else if (text.StartsWith(FINISHED_WORKING))
				{
					buildFinished = true;
					shouldBeStandingStill = false;
				}

				break;
			}
		case MSGTYPE_PERSIST_ITEM:
			{
				if (actionToDo == ACTION_BUILD || actionToDo == ACTION_READRECIPE || actionToDo == ACTION_DROP_TABLE || actionToDo == ACTION_SCAN_TABLE || actionToDo == ACTION_TEST_TABLE || actionToDo == ACTION_EXPLORE)
				{
					psPersistItem incoming(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
					csVector3 currPos = worldHandler::GetCurrentPos();
					float itemDistance = sqrt(csSquaredDist::PointPoint(currPos, incoming.pos));

					if (incoming.name.EndsWith(CRAFT_TABLE) && waitingForDrop && itemDistance < 3.0f)
					{
						tableEid = incoming.eid;
						tablePos = incoming.pos;
						if (needTableScan)
						{
							worldHandler::TargetEID(tableEid);
							worldHandler::OpenTargetEID(tableEid);
							prevRotation = worldHandler::GetCurrentRot();
						}
						else if (needTableTest)
						{
							DoSendTableTest();
							worldHandler::SetRotation(prevRotation);
						}
						waitingForDrop = false;
					}
					else if (incoming.name.EndsWith(EXP_MARKER) && itemDistance <= 5.0f)
					{
						if (!possibleClaims.Contains(incoming.eid)) {
							possibleClaims.Put(incoming.eid, incoming);
						}
					}
					else if (incoming.name.StartsWith(ITEM_NAME_FOR_TINY_CLAIMS) && itemDistance <= 5.0f)
					{
						lootBags.PutUnique(incoming.eid, incoming.eid);
					}
				}
				break;
			}
		}
	message->Reset();
}
csArray<move> ExploreActivity::Stack(csArray<slot> invSlots)
{
	slot sample = invSlots.Get(0);
	// find all the slots with matching name in this quality band
	csArray<slot> tableSlots = FindTableSlots(sample.name, sample.quality, qualityBand);
	tableSlots.Merge(invSlots);
	tableSlots.Sort(); // slots sort on quality
	printf("merged slots sorted on quality:\n");
	printSlots(tableSlots);
	csArray<move> result = csArray<move>();
	move nextMove = ComputeNextMove(tableSlots);
	while (nextMove.count > 0)
	{
		result.Push(nextMove);
		tableSlots = SimulateMove(nextMove, tableSlots);
		nextMove = ComputeNextMove(tableSlots);
	}
	// make sure all the inv slots made it into the table
	size_t nextEmptySlot = 0;
	for (size_t i = 0; i < tableSlots.GetSize(); i++)
	{
		if (tableSlots.Get(i).isInInventory)
		{
			int emptySlot = -1;
			for (int j = nextEmptySlot; j < 16; j++)
			{
				if (tableContents[j].count == 0)
				{
					emptySlot = j;
					nextEmptySlot = j+1;
					// reserve this (soon to be occupied) slot
					tableContents[emptySlot].count = -1;
					break;
				}
			}
			if (emptySlot == -1)
			{
				OutputMsg(csString("Table is full. Stopping."));
				noBuild = true;
				exploreFinished = true;
				actionToDo = ACTION_NONE;
				DoStop();
				Stop();
				return csArray<move>();
			}
			tableContents[emptySlot].name = tableSlots.Get(i).name;
			tableContents[emptySlot].quality = tableSlots.Get(i).quality;

			move newMove(tableSlots.Get(i), tableContents[emptySlot], tableSlots.Get(i).count);
			printMove(newMove);
			result.Push(newMove);
		}
	}
	printf("done stacking\n");
	return result;
}
csArray<slot> ExploreActivity::SimulateMove(move m, csArray<slot> tableSlots)
{
	printf("\n");
	printMove(m);
	csArray<slot> result = csArray<slot>(tableSlots);
	int x = m.count;
	int y = m.dest.count;
	int q1 = m.src.quality;
	int q2 = m.dest.quality;
	int outputQual = (x*q1 + y*q2) / (x+y);
	slot srcAfter = slot(m.src.slotNo, m.src.name, m.src.count - m.count, m.src.quality, m.src.isInInventory);
	slot destAfter = slot(m.dest.slotNo, m.dest.name, m.dest.count + m.count, outputQual, m.dest.isInInventory);
	result.Delete(m.src);
	result.Delete(m.dest);
	result.Push(destAfter);
	if (srcAfter.count > 0)
		result.Push(srcAfter);
	result.Sort();
	printSlots(result);
	return result;
}
void ExploreActivity::printMove(move m)
{
	csString out = csString(m.src.isInInventory ? "inv:" : "tab:").Append(m.src.slotNo).Append(" x").Append(m.count).Append(" --> ");
	out.Append(m.dest.isInInventory ? "inv:" : "tab:").Append(m.dest.slotNo);
	//OutputMsg(out);
	printf(out.GetData());
	printf("\n");
}
void ExploreActivity::printSlots(csArray<slot> slots)
{
	csString out = csString("");
	for (size_t i = 0; i < slots.GetSize(); i++)
	{
		slot s = slots.Get(i);
		out.Append(s.isInInventory ? "inv:" : "tab:").Append(s.slotNo).Append("[").Append(s.count).Append("x ").Append(s.quality).Append("q] ");
	}
	//OutputMsg(out);
	printf(out.GetData());
	printf("\n");
}
move ExploreActivity::ComputeNextMove(csArray<slot> slots)
{
	if (slots.GetSize() < 2)
		return move(slot(), slot(), 0);
	// look for same quality pair
	for (size_t i = 0; i < slots.GetSize()-1; i++)
	{
		if (slots.Get(i).quality == slots.Get(i+1).quality)
		{
			return BuildMove(slots.Get(i), slots.Get(i+1));
		}
	}
	// build all combos of odd-odd and even-even quality pairs
	csArray<slot> odds = csArray<slot>();
	csArray<slot> evens = csArray<slot>();
	for (size_t i = 0; i < slots.GetSize(); i++)
	{
		if (slots.Get(i).quality % 2 == 0)
			evens.Push(slots.Get(i));
		else
			odds.Push(slots.Get(i));
	}
	csArray<slotpair> pairs = Combinations(odds);
	pairs.Merge(Combinations(evens));
	if (pairs.IsEmpty())
		return move(slot(), slot(), 0);
	pairs.Sort();
	return BuildMove(pairs.Get(0).a, pairs.Get(0).b);
}
move ExploreActivity::BuildMove(slot a, slot b)
{
	// if can move either way, move out of inv and left
	if (a.quality == b.quality || a.count == b.count)
	{
		if (a.isInInventory || (!b.isInInventory && a.slotNo > b.slotNo))
			return move(a, b, a.count);
		return move(b, a, b.count);
	}
	// must move from high count to low count
	if (a.count > b.count)
		return move(a, b, b.count);
	return move(b, a, a.count);
}
csArray<slotpair> ExploreActivity::Combinations(csArray<slot> slots)
{
	csArray<slotpair> result = csArray<slotpair>();
	if (slots.GetSize() >= 2)
	{
		for (size_t i = 0; i < slots.GetSize()-1; i++)
		{
			for (size_t j = i+1; j < slots.GetSize(); j++)
			{
				result.Push(slotpair(slots.Get(i), slots.Get(j)));
			}
		}
	}
	return result;
}
void ExploreActivity::DoSendTableTest()
{
	needTableTest = false;
	int slotToTest = -1;
	for (int i = 0; i < 16; i++)
	{
		if (tableContents[i].count > 0)
		{
			slotToTest = i;
			break;
		}
	}
	if (slotToTest == -1)
	{
		OutputMsg(csString("Testing table eid: ").Append(tableEid.Unbox()));
		OutputMsg(csString("    can't test empty table"));
		if (tableEid.Unbox() > 1000)
		{
			tableTestPassed = true;
			OutputMsg(csString("    presuming empty table good based on eid"));
			printf("table presumed good based on eid:%d\n", tableEid.Unbox());
		}
		else
		{
			tableTestPassed = false;
			OutputMsg(csString("    presuming empty table bad based on eid"));
			printf("table presumed bad based on eid:%d\n", tableEid.Unbox());
		}
	}
	else 
	{
		tableTestSent = true;
		OutputMsg(csString("Testing table eid: ").Append(tableEid.Unbox()));
		tableTestStart = csGetTicks();
		worldHandler::MoveItems(tableEid.Unbox(), slotToTest, tableEid.Unbox(), slotToTest, 1);
	}
}
csArray<slot> ExploreActivity::FindTableSlots(csString name, int quality, int band)
{
	csArray<slot> result;
	for (int i = 0; i < 16; i++)
	{
		if (tableContents[i].count > 0)
		{
			if(tableContents[i].name.Compare(name))
			{
				if ((tableContents[i].quality / band) == (quality / band))
				{
					result.Push(tableContents[i]);
				}
			}
		}
	}
	return result;
}
void ExploreActivity::ScheduleNewPerform(int delay)
{
	csRef<iEventTimer> timer = worldHandler::GetStandardTimer();
	timer->AddTimerEvent(this, delay);
}
bool ExploreActivity::Perform(iTimerEvent *ev)
{

	if (!IsOngoing() && !IsFinished())
		return false;

	switch (actionToDo)
	{
		case ACTION_INIT:
		{
			DoInit();
			NextAction(MIN_DELAY);
			break;
		}
		case ACTION_DROP_TABLE:
		{
			DoDropTable();
			NextAction(3*MIN_DELAY);
			break;
		}
		case ACTION_SCAN_TABLE:
		{
			DoScanTable();
			NextAction(MIN_DELAY);
			break;
		}
		case ACTION_TEST_TABLE:
		{
			DoTestTable();
			NextAction(5*MIN_DELAY);
			break;
		}
		case ACTION_EXPLORE:
		{
			DoExplore();
			NextAction(maxDelay);
			break;
		}
		case ACTION_READRECIPE:
		{
			DoReadRecipe();
			NextAction(3*MIN_DELAY);
			break;
		}
		case ACTION_BUILD:
		{
			if (exploreFinished && !Compare(markerPos, csVector3(0.0f,0.0f,0.0f)))
			{
				if (DoBuild())
					NextAction(3*MIN_DELAY);
				else return true;	//try again
			}
			else if (csGetTicks() - startTime < timeout)
			{
				OutputMsg("Explore not finished and not yet timedout, will wait.");
				return true;
			}
			else
			{
				OutputMsg("Explore not finished, but timed out, will move on!");
				noBuild = true;
				NextAction(MIN_DELAY);
			}
			break;
		}
		case ACTION_USE_COMBINE:
		{
			DoUseCombine();
			NextAction(maxDelay);
			break;
		}
		case ACTION_EXAMINE_RESULT:
		{
			DoExamineResult();
			NextAction(3*MIN_DELAY);
			break;
		}
		case ACTION_TAKE_RESULT:
		{
			DoTakeResult();
			NextAction(2*MIN_DELAY);
			break;
		}
		case ACTION_DIE:
		{
			DoDie();
			NextAction(DEATH_DELAY);
			break;
		}
		case ACTION_DROP_KEY:
		{
			DoDropKey();
			NextAction(2*MIN_DELAY);
			break;
		}
		case ACTION_STORE_RESOURCES:
		{
			DoStoreResources();
			NextAction(MIN_DELAY);
			break;
		}
		case ACTION_MOVE:
		{
			DoMove();
			NextAction(MIN_DELAY);
			break;
		}
		case ACTION_UNMOVE:
		{
			DoUnMove();
			NextAction(MIN_DELAY);
			break;
		}
		case ACTION_MOVE_TOWARD_TABLE:
		{
			DoMoveTowardTable();
			NextAction(MIN_DELAY);
			break;
		}

		case ACTION_STOP:
		{
			DoStop();
			NextAction(2*MIN_DELAY);
			break;
		}
		case ACTION_NONE:
		{
			Finish();
			break;
		}
	}
	return false;
}

void ExploreActivity::StartWork()
{
	OutputMsg(csString("Explore activity started."));
	ScheduleNewPerform(MIN_DELAY);
	timesDone = 0;
}
void ExploreActivity::WrapUp()
{
	csVector3 pos = worldHandler::GetCurrentPos();
	csString sep = csString(",");
	char x[100], y[100], z[100];
	sprintf(x, "%.2f", pos.x);
	sprintf(y, "%.2f", pos.y);
	sprintf(z, "%.2f", pos.z);
	csString toolName = worldHandler::GetEquippedToolName();

	csHash<int64_t, csString>::GlobalIterator items = claimOutput.GetIterator();
	while(items.HasNext())
	{
		csString itemName;
		int64_t count = items.Next(itemName);
		char itemCount[100];
		sprintf(itemCount, "%d", count);

		csString text = toolName +sep+csString(x)+sep+csString(y)+sep+csString(z)+sep+claimName+sep+csString(itemCount)+sep+itemName;
		if (logger == NULL)
			OutputMsg(csString("No logger available, so activities will NOT be logged."));
		else
			logger->Write(CSV_EXPLORING, text);
	}
	claimOutput.DeleteAll();
	if (timesDone > timesToRepeat)
	{
		OutputMsg(csString("Explore activity finished."));
	}
	else
	{
		char msg[1000];
		sprintf(msg, "%d explore attempts done, %d attempts left.", timesDone, timesToRepeat-timesDone);
		OutputMsg(csString(msg));
	}
}
void ExploreActivity::DoInit()
{
	actionToDo = ACTION_INIT;
	attemptsCount = 0;
	claimName = csString("");
	scrollName = csString("");
	claimType = csString("Unknown");
	keyName = csString("");
	markerID = EID(-1000);
	tableEid = EID(-1001);
	lastKnownTableDist = -1.0f;

	markerWaitCount = 0;
	preResultInvVersion = 0;
	itemName = csString("");

	needToDie = false;
	noBuild = true;
	exploreFinished = false;
	buildFinished = false;
	needTableScan = false;
	needTableTest = false;
	tableTestSent = false;
	tableTestPassed = false;
	seekingTableUrgently = false;
	glitchPickup = false;
	parseClaimKey = false;
	verifyClaimTransfer = false;
	claimTransferSuccess = true;
	tableScanSent = false;
	tableScanInProgress = false;
	storingResourcesComplete = false;
	storingResourcesInProgress = false;
	verifyInventoryMove = false;
	verifyTableSlotMove = false;
	shouldBeStandingStill = true;
	waitForScroll = false;
	rebuild = false;
	pickupGroundLoot = false;
	markerPos = csVector3(0.0f,0.0f,0.0f);
	pendingResourceMoves.DeleteAll();
	claimRcp.ParseRecipe("", false);
	waitingForDrop = false;

	startTime = csGetTicks();
	tableTestTimeout = csTicks(5000);

	int sz;
	EXPLORE_STRATEGIES strat;
	if (IsOngoing() && pickupDone)
	{
		worldHandler::Nuke(6000);
	}

	if (timesDone == 0 && IsOngoing())
	{
		csString initMsg = csString("  ").Append(timesToRepeat).Append("x ");
		toolSlot = NULL;
		tableSlot = -1;

		WordArray words(cmd,false);

		if (words.GetCount()>0)
		{
			csString s;
			words.GetString(0, s);
			if (s.CompareNoCase(csString("line")))
			{
				strat = EXPLORE_LINE;
				initMsg = initMsg.Append("line");
			}
			else if (s.CompareNoCase(csString("grid")))
			{
				strat = EXPLORE_GRID;
				initMsg = initMsg.Append("grid");
			}
			else
			{
				strat = EXPLORE_LINE;
				initMsg = initMsg.Append("line");
			}
		}
		else
		{
			strat = EXPLORE_LINE;
			initMsg = initMsg.Append("line");
		}

		if (words.GetCount()>1)
			sz = words.GetInt(1);
		else sz = DEFAULT_SIZE;

		if (sz == 0)
		{
			sz = DEFAULT_SIZE;
		}
		initMsg = initMsg.Append(" size-").Append(sz);

		if (words.GetCount() > 2)
			if (words.GetInt(2) == 0)
				{
					leaveKeys = false;
					initMsg = initMsg.Append(" keep-keys ");
				}
			else
				{
					leaveKeys = true;
					initMsg = initMsg.Append(" drop-keys ");
				}

		if (words.GetCount() > 3 && words[3] == "m")
		{
			useMax = false;
			initMsg = initMsg.Append(" min-mats ");
		}
		else 
		{
			useMax = true;
			initMsg = initMsg.Append(" max-mats ");
		}

		if (words.GetCount() > 4)
			maxDelay = words.GetInt(4);
		else
			{
				maxDelay = MAX_DELAY;
			}

		if (words.GetCount() > 5)
			timeout = words.GetInt(5);
		else
			{
				timeout = DEFAULT_EXPLORE_TIMEOUT;
			}
		strategy.Init(strat, sz, startPoint, startYRot);

		if (words.GetCount() > 6)
			qualityBand = words.GetInt(6);
		else
			qualityBand = 1;
		initMsg = initMsg.Append(" quality-band=").Append(qualityBand).Append(" ");

		initialTableCount = CountTables();
		tableSlot = FindTableSlot();
		usingTable = tableSlot > 0;
		if (usingTable)
		{
			needTableScan = true;
			needTableTest = false;
			initMsg = initMsg.Append("with-table");
			for (int i = 0; i < 16; i++)
				emptyTableSlots.Add(i);
		}
		else
		{
			OutputMsg(csString("NO TABLE WAS FOUND IN INVENTORY"), TEXT_COLOUR_WHITE);
			initMsg = initMsg.Append("no-table");
		}

		OutputMsg(initMsg);
	}
}

int ExploreActivity::SelectToolSlot(csString tool, int quality)
{
	csArray<psInventoryCache::CachedItemDescription*>::Iterator iter = exploreInvCache->GetSortedIterator();
	while (iter.HasNext())
	{
	    psInventoryCache::CachedItemDescription* slotDesc = iter.Next();
		if (slotDesc->name.CompareNoCase(tool))
			if (slotDesc->itemQuality > quality) 
				return slotDesc->slot;
	}
	return -1;
}
int ExploreActivity::FindTableSlot()
{
	csArray<psInventoryCache::CachedItemDescription*>::Iterator iter = exploreInvCache->GetSortedIterator();
	tableSlot = -1;
	float maxTableWeight = 0.0f;
	int tablesFound = 0;
	while (iter.HasNext())
	{
	    psInventoryCache::CachedItemDescription* slotDesc = iter.Next();
		if (slotDesc->name.CompareNoCase(CRAFT_TABLE))
	    {
			tablesFound++;
			if (slotDesc->weight > maxTableWeight) 
			{	
				tableSlot = slotDesc->slot;
				maxTableWeight = slotDesc->weight;
			}
	    }
	}
	if (tablesFound < initialTableCount) return -1;
	return tableSlot;
}
int ExploreActivity::CountTables()
{
	csArray<psInventoryCache::CachedItemDescription*>::Iterator iter = exploreInvCache->GetSortedIterator();
	int count = 0;
	while (iter.HasNext())
	{
	    psInventoryCache::CachedItemDescription* slotDesc = iter.Next();
		if (slotDesc->name.CompareNoCase(CRAFT_TABLE))
			count++;
	}
	return count;
}

void ExploreActivity::NextAction(int delay)
{
	switch (actionToDo)
	{
		case ACTION_INIT:
		{
			if (timesDone >= timesToRepeat)
				{
					actionToDo = ACTION_NONE;
				}
			else if (usingTable)
				actionToDo = ACTION_DROP_TABLE;
			else actionToDo = ACTION_EXPLORE;
			break;
		}
		case ACTION_DROP_TABLE:
		{
			if (tableEid.Unbox() == -1001)
				actionToDo = ACTION_DROP_TABLE;
			else if (needTableScan)
			{
				tableScanInProgress = true;
				actionToDo = ACTION_SCAN_TABLE;
			}
			else
				actionToDo = ACTION_TEST_TABLE;
			break;
		}
		case ACTION_SCAN_TABLE:
		{
			if (tableScanInProgress)
				actionToDo = ACTION_SCAN_TABLE;
			else
				actionToDo = ACTION_TEST_TABLE;
			break;
		}
		case ACTION_TEST_TABLE:
		{
			if (glitchPickup)
				actionToDo = ACTION_DROP_TABLE;
			else if (!tableTestPassed)
				actionToDo = ACTION_TEST_TABLE;
			else
				actionToDo = ACTION_EXPLORE;
			break;
		}
		case ACTION_EXPLORE:
		{
			if (needToDie)
				actionToDo = ACTION_DIE;
			else actionToDo = ACTION_READRECIPE;
			break;
		}
		case ACTION_READRECIPE:
		{
			if (needToDie)
				actionToDo = ACTION_DIE;
			else if (waitForScroll)
				actionToDo = ACTION_READRECIPE;
			else actionToDo = ACTION_BUILD;
			break;
		}
		case ACTION_BUILD:
		{
			if (noBuild)
				if (needToDie)
					actionToDo = ACTION_DIE;
				else actionToDo = ACTION_DROP_KEY;
			else
				actionToDo = ACTION_USE_COMBINE;
			break;
		}
		case ACTION_USE_COMBINE:
		{
			if (!claimTransferSuccess)
				actionToDo = ACTION_DROP_KEY;
			else
				actionToDo = ACTION_EXAMINE_RESULT;
			break;
		}
		case ACTION_EXAMINE_RESULT:
		{
			if (buildFinished)
				actionToDo = ACTION_TAKE_RESULT;
			else if (rebuild)
			{
				actionToDo = ACTION_USE_COMBINE;
			}
			else
			{
				OutputMsg(csString("    waiting for build to finish..."));
				actionToDo = ACTION_EXAMINE_RESULT;
			}
			break;
		}
		case ACTION_TAKE_RESULT:
		{
			if (needToDie)
				actionToDo = ACTION_DIE;
			else if (watchingInvForNewResources)
				actionToDo = ACTION_TAKE_RESULT;
			else actionToDo = ACTION_DROP_KEY;
			break;
		}
		case ACTION_DIE:
		{
			actionToDo = ACTION_EXPLORE;
			break;
		}
		case ACTION_DROP_KEY:
		{
			if (usingTable && !noBuild)
				actionToDo = ACTION_STORE_RESOURCES;
			else actionToDo = ACTION_MOVE;
			break;
		}
		case ACTION_STORE_RESOURCES:
		{
			if (watchingInvForNewResources || storingResourcesInProgress || !storingResourcesComplete)
				actionToDo = ACTION_STORE_RESOURCES;
			else
				actionToDo = ACTION_MOVE;
			break;
		}
		case ACTION_MOVE:
		{
			actionToDo = ACTION_STOP;
			break;
		}
		case ACTION_UNMOVE:
		{
			actionToDo = ACTION_STOP;
			break;
		}
		case ACTION_MOVE_TOWARD_TABLE:
		{
			actionToDo = ACTION_STOP;
			break;
		}
		case ACTION_STOP:
		{
			actionToDo = ACTION_INIT;
			break;
		}
		default:
		{
			actionToDo = ACTION_NONE;
			return;
		}
	}
	ScheduleNewPerform(delay);
}

void ExploreActivity::DoDropTable()
{
	if (tableEid.Unbox() != -1001)
		return;
	int tableSlot = FindTableSlot();
	if (tableSlot == -1)
	{
		OutputMsg(csString("looks like table has already been droppped"));
		return;
	}
	OutputMsg(csString("Dropping table from slot ").Append(tableSlot));
	if (!needTableScan)
	{
		needTableTest = true;
	}
	worldHandler::DropFromSlot(tableSlot, 1);
	waitingForDrop = true;
}

void ExploreActivity::DoScanTable()
{
	if (tableScanInProgress && !tableScanSent) 
	{
		worldHandler::OpenTargetEID(tableEid);
		tableScanSent = true;
	}
}

void ExploreActivity::DoTestTable()
{
	if (!tableTestPassed)
	{
		if (csGetTicks() - tableTestStart < tableTestTimeout)
		{
			OutputMsg(csString("    table looks bad, could be lag...waiting"));
			return;
		}

		tableTestSent = false;
		glitchPickup = true;
		needTableTest = true;
		OutputMsg(csString("    re-dropping glitched table"));
		PickupTable();
	}
	else
	{
		OutputMsg(csString("    table works"));
	}
}

void ExploreActivity::DoExplore()
{
	OutputMsg(csString("Exploring..."));
	nextFreeSlot = worldHandler::GetNextFreeInvSlot();
	worldHandler::ExecCmd(csString("/explore"));
}
void ExploreActivity::DoReadRecipe()
{
	if (needToDie)
	{
		OutputMsg(csString("    need to die"));
		return;
	}

	if (!scrollName.CompareNoCase("")) {
		if (itemName.CompareNoCase(claimRcp.GetItemName())) // parsed recipe matches scroll
		{
			waitForScroll = false;
			return;
		}
		psInventoryCache::CachedItemDescription* from = worldHandler::FindItemSlot(scrollName, true);
		if (from)
		{
			psViewItemDescription out(from->containerID, from->slot);
			out.SendMessage();
			if (waitForScroll) 
				OutputMsg(csString("    waiting for recipe ingredient list"));
		}
		else
		{
			OutputMsg(csString("    waiting for enum to appear in inv"));
		}
	}
	else
	{
		OutputMsg(csString("    waiting for 'You got Enum' msg"));
	}
	waitForScroll = true;
}

bool ExploreActivity::DoBuild()
{
	csHash<psPersistItem, EID>::GlobalIterator markers = possibleClaims.GetIterator();
	while(markers.HasNext())
	{
		psPersistItem candidate = markers.Next();
		csVector3 pos = candidate.pos;
		if (Compare(candidate.pos, markerPos))
		{
			markerID = candidate.eid;
			worldHandler::TargetEID(markerID);
			OutputMsg(csString("    found claim marker eid: ").Append(markerID.Unbox()));
			printf("*selected* claim eid: %d (%f, %f, %f)\n", candidate.eid.Unbox(), pos.x, pos.y, pos.z);
			break;
		}
		else
		{
			printf("rejecting claim eid: %d (%f, %f, %f)\n", candidate.eid.Unbox(), pos.x, pos.y, pos.z);
		}
	}
	possibleClaims.DeleteAll();
	if (markerID.Unbox() == -1000)
	{
		if (markerWaitCount < 2)
		{
			OutputMsg(csString("    no claim marker yet.. waiting"));
			printf("    no claim marker yet.. waiting\n");
		}
		else if (markerWaitCount % 2 == 0)
		{
			OutputMsg(csString("    stepping toward marker"));
			printf("stepping toward marker\n");
			worldHandler::StepToward(markerPos);
			psengine->GetCelClient()->GetMainPlayer()->StopMoving(true);
			worldHandler::StopAllMovement();
		}
		else
		{
			OutputMsg(csString("    still waiting..."));
			printf("still waiting...\n");
		}
		markerWaitCount++;
		return false;
	}

	psengine->GetCelClient()->GetMainPlayer()->StopMoving(true);
	worldHandler::StopAllMovement();
	worldHandler::SendDRUpdate();
	markerWaitCount = 0;

	if (!noBuild && !itemName.CompareNoCase("") && markerID.Unbox() != -1000 && !scrollName.CompareNoCase("") && !claimName.CompareNoCase(""))
	{
		worldHandler::TargetEID(markerID);
		csString targetName = worldHandler::GetTargetName();
		worldHandler::OpenTargetEID(markerID);
		int toContainer = markerID.Unbox();

		if (!worldHandler::IsInBrain(scrollName))
		{
			if (!worldHandler::EquipRecipe(scrollName))
		         {
					OutputMsg(csString("Missing blueprint for the claim!"));
					noBuild = true;
					//Error();
		            return false;
		         }
		}

		OutputMsg(csString("Building claim with:"));
		psInventoryCache::CachedItemDescription* from = worldHandler::FindItemSlot(claimRcp.GetBundleName(), true);
		if (from)
		{
			worldHandler::MoveItems(from->containerID, from->slot, toContainer, 3, 1);
			verifyClaimTransfer = true;
			OutputMsg(csString("    1x ").Append(claimRcp.GetBundleName()));
			return true;
		} 
		else if (usingTable) 
		{
			// TODO: check the table for existing bundles
		}

		csHash<int, csString>::GlobalIterator iterIngredients(claimRcp.GetIngredientsList()->GetIterator());

		int nextEmptySlot = 3;
		if (!iterIngredients.HasNext())
		{
			OutputMsg(csString("Empty ingredients list! Can't build."));
			noBuild = true;
			Error();
			return true;
		}
		while (iterIngredients.HasNext())
		{
			csString itemName;
			int quantity = iterIngredients.Next(itemName);
			int slot = SelectIngredientSlot(itemName, quantity);

			if (slot < 0) 
			{
				OutputMsg(csString("Not enough ").Append(itemName).Append("!"));
				// reclaim any previously moved mats in claim
				worldHandler::TakeAllFromTarget();
				verifyClaimTransfer = false;
				noBuild = true;
				if (claimType.CompareNoCase("Tiny") || claimType.CompareNoCase("Small"))
				{
					OutputMsg(csString("Bot Stopping"));
					Error();
				}
				else
				{
					OutputMsg(csString("    Locking and moving on"));
					worldHandler::PicklockEID(markerID);
				}
				return true;
			}
			worldHandler::MoveItems(CONTAINER_INVENTORY_EQUIPMENT, slot, toContainer, nextEmptySlot++, quantity);
			verifyClaimTransfer = true;
			OutputMsg(csString("    ").Append(quantity).Append("x ").Append(itemName));
		}
		worldHandler::TargetEID(markerID);
	}
	else
		noBuild = true;
	return true;
}

int ExploreActivity::SelectIngredientSlot(csString name, int quantity)
{
	csArray<psInventoryCache::CachedItemDescription*>::Iterator iter = exploreInvCache->GetSortedIterator();
	while (iter.HasNext())
	{
	    psInventoryCache::CachedItemDescription* slotDesc = iter.Next();
		if (slotDesc->name.CompareNoCase(name) && slotDesc->stackCount >= quantity)
			return slotDesc->slot;
	}
	return -1;
}

void ExploreActivity::DoUseCombine()
{
	verifyClaimTransfer = false;
	if (!claimTransferSuccess)
	{
		printf("claimTransfer failed for eid: %d\n", markerID.Unbox());
		noBuild = true;
		return;
	}
	
	printf("claimTransfer succeeded for eid: %d\n", markerID.Unbox());
	if (rebuild)
	{
		OutputMsg(csString("    trying again since you moved"));
		rebuild = false;
	}
	worldHandler::TargetEID(markerID);
	if (claimType.CompareNoCase("Tiny"))
		worldHandler::UseTarget();
	else
		worldHandler::CombineContentsInTarget();
}
void ExploreActivity::DoExamineResult()
{
	shouldBeStandingStill = false;
}

void ExploreActivity::DoTakeResult()
{
	if (watchingInvForNewResources)
	{
		OutputMsg(csString("    waiting for /takeall..."));
		printf("inventory not showing results of /takeall. waiting more...\n");
		return;
	}

	if (preResultInvVersion != 0) return;
	OutputMsg(csString("Taking output..."));
	preResultInvVersion = exploreInvCache->GetInventoryVersion();
	invSnapshot = InventorySnapshot();
	worldHandler::TakeAllFromTarget();
	watchingInvForNewResources = true;
}

csBitArray ExploreActivity::InventorySnapshot()
{
	csBitArray result(80);
	for (int i = 0; i < 80; i++)
	{
		psInventoryCache::CachedItemDescription* slot = exploreInvCache->GetInventoryItem(i+16);
		if (slot != NULL)
		{
			result.Set(i);
		}
	}
	return result;
}

void ExploreActivity::DoStoreResources()
{
	// make sure newer exploreInvCache version
	if (exploreInvCache->GetInventoryVersion() == preResultInvVersion)
	{
		OutputMsg(csString("    waiting for /takeall..."));
		printf("inventory not showing results of /takeall. waiting more...\n");
		return;
	}

	if (storingResourcesComplete) return;

	if (storingResourcesInProgress) return;

	if (csGetTicks() - lastVerifyTime < MIN_DELAY) return;

	storingResourcesInProgress = true;
	ProcessNextResourceMove();
}

void ExploreActivity::ProcessNextResourceMove()
{
	if (pendingResourceMoves.IsEmpty())
	{
		storingResourcesComplete = true;
		storingResourcesInProgress = false;
		return;
	}
	currentResourceMove = pendingResourceMoves.Front();
	pendingResourceMoves.PopFront();
	if (currentResourceMove.dest.count == -1)
		currentResourceMove.dest.count = 0;
	csString moveMsg = csString("      ").Append(currentResourceMove.count).Append("x ").Append(currentResourceMove.src.name);
	moveMsg.Append(currentResourceMove.src.isInInventory ? " inventory[" : " table[").Append(currentResourceMove.src.slotNo);
	moveMsg.Append("] --> ").Append(currentResourceMove.dest.isInInventory ? " inventory[" : " table[").Append(currentResourceMove.dest.slotNo);
	moveMsg.Append("]");
	OutputMsg(moveMsg);
	int fromContainer = currentResourceMove.src.isInInventory ? -2 : tableEid.Unbox();
	int toContainer = tableEid.Unbox();
	if (currentResourceMove.dest.isInInventory)
	{
		verifyInventoryMove = true;
		toContainer = -2;
	}
	else
	{
		verifyTableSlotMove = true;
	}
	worldHandler::MoveItems(fromContainer, currentResourceMove.src.slotNo, 
		toContainer, currentResourceMove.dest.slotNo, currentResourceMove.count);
}

void ExploreActivity::DoDropKey()
{
	//move key to container
	if (leaveKeys)
	{
		if (!claimTransferSuccess || markerID.Unbox() == -1000)
		{
			OutputMsg(csString("    move into claim marker failed."));
			if (claimType.CompareNoCase("Tiny") || claimType.CompareNoCase(""))
			{
				psSlotMovementMsg msg( nextFreeSlot->ContainerID(), nextFreeSlot->ID(), CONTAINER_WORLD, 0, 1);
				msg.SendMessage();
				OutputMsg(csString("    dropping key for glitched tiny claim."));
				printf("Move item into claim marker failed. Dropping key.\n");
			}
			else
			{
				OutputMsg(csString("    locking glitched claim."));
				worldHandler::PicklockEID(markerID);
			}

		}
		else if ((claimType.CompareNoCase("Tiny") || claimType.CompareNoCase("Small")) && nextFreeSlot!=NULL && nextFreeSlot->GetToolTip().Find("Key") < nextFreeSlot->GetToolTip().Length())
		{
			OutputMsg(csString("Dropping key"));
			int container   = nextFreeSlot->ContainerID();
			int slot        = nextFreeSlot->ID();
			psSlotMovementMsg msg( container, slot, markerID.Unbox(), 3, 1);
			msg.SendMessage();
		}
		else
		{	// explain why not leaving the key
			printf("Not dropping key because: ");
			if (!(claimType.CompareNoCase("Tiny") || claimType.CompareNoCase("Small"))) printf("claimType '%s' is not tiny or small. ", claimType.GetData());
			if (nextFreeSlot == NULL) printf("next free slot is null. ");
			if (nextFreeSlot->GetToolTip().Find("Key") >= nextFreeSlot->GetToolTip().Length())
			{
				printf("nextFreeSlot doesn't contain a key old:'%s'\n", nextFreeSlot->GetToolTip().GetData());
				csArray<psInventoryCache::CachedItemDescription*>::Iterator iter = exploreInvCache->GetSortedIterator();
				while (iter.HasNext())
				{
					psInventoryCache::CachedItemDescription* slotDesc = iter.Next();
					printf("  %2d: %s\n", slotDesc->slot-16, slotDesc->name.GetData());
				}
			}
			printf("\n");
		}
	}

	worldHandler::TargetEID(tableEid);
	worldHandler::OpenTargetEID(tableEid);
}

void ExploreActivity::DoDie()
{
	OutputMsg(csString("Dying..."));
	worldHandler::ExecCmd(csString("/die"));
	needToDie = false;
}
void ExploreActivity::DoMove()
{
	if (lootBags.GetSize() > 0)
	{
		OutputMsg(csString("Picking up lbn x").Append(lootBags.GetSize()));
		csHash<EID, EID>::GlobalIterator lootList(lootBags.GetIterator());

		while (lootList.HasNext()) {
			worldHandler::PickupEID(lootList.Next());
		}
		lootBags.DeleteAll();
	}

	if (pickupGroundLoot)
	{
		OutputMsg(csString("Loot on ground, bot stopping."));
		actionToDo = ACTION_NONE;
		DoStop();
		Stop();
		return;
	}
	OutputMsg(csString("Moving"));
	strategy.Move(timesDone);
}
void ExploreActivity::DoUnMove()
{
	strategy.UnMove(timesDone);
}
void ExploreActivity::DoMoveTowardTable()
{
	worldHandler::TurnToPoint(tablePos.x, tablePos.z);
	worldHandler::StepForward();
}
void ExploreActivity::DoStop()
{
	worldHandler::StopAllMovement();
	worldHandler::SendDRUpdate();
	shouldBeStandingStill = true;
	bool pickupAttempted = false;
	if (usingTable)
		pickupAttempted = PickupTable();
		
	if (!pickupAttempted)	
		Finish();
	// else wait for table to arrive in inventory to call Finish()
}
bool ExploreActivity::PickupTable()
{
	if (tableEid.Unbox() == -1001) {
		OutputMsg(csString("tableEid == -1001"));
		GEMClientObject* tableObj = worldHandler::FindItemInWorld(CRAFT_TABLE);
		if (tableObj != NULL)
		{
			tableEid = tableObj->GetEID();
		}
	}
	if (tableEid.Unbox() != -1001) {
		OutputMsg(csString("Picking up table"));
		worldHandler::StopAllMovement();
		worldHandler::SendDRUpdate();
		prevRotation = worldHandler::GetCurrentRot();
		pickupDone = false;
		return worldHandler::PickupEID(tableEid);
	}
	else OutputMsg(csString("Table with resources NOT FOUND in world!"));
	return false;
}

/*************************Craft Activity*****************/

CraftActivity::CraftActivity()
{
	name = csString("craft");
	textcolour = TEXT_COLOUR;
	craftInvCache = new psInventoryCache();
	watchingInvForNewResources = false;
}
CraftActivity::~CraftActivity()
{

}
void CraftActivity::StartWork()
{
	OutputMsg(csString("Craft activity started."));
	ScheduleNewPerform(MIN_DELAY);
}
void CraftActivity::WrapUp()
{
	if (logger == NULL)
		OutputMsg(csString("No logger available, so activities will NOT be logged."));
	else
		logger->Write(CSV_CRAFTING, logText);
}
void CraftActivity::HandleMessage(MsgEntry* message )
{
	if (!IsOngoing())
		return;

	switch (message->GetType() )
	{
		case MSGTYPE_MODE:
		{
			break;
		}
		case MSGTYPE_GUISKILL:
		{
			craftFinished = true;
			break;
		}
		case MSGTYPE_SYSTEM:
		{
			psSystemMessage incoming(message);
			csString text = incoming.msgline;

			if (text.StartsWith(LOOT_ON_GROUND))
			{
    				lootOnGround = true;
			}
			else if (text.StartsWith(RANKING_PREFIX))
			{
				needToTrain = true;
				size_t end = text.FindLast('.');

				size_t start = RANKING_PREFIX.Length();
				skillToTrain = text.Slice(start, end-start);
				OutputMsg(csString("    ranked up in ")+skillToTrain);
			}
			else if (text.StartsWith(CONTAINER_WORN))
			{
				Error();
				OutputMsg("Container worn-out, crafting will stop.");
			}
			else if (text.StartsWith(TOOL_WORN))
			{
				Error();
				OutputMsg("Tool worn-out, crafting will stop.");
			}
			else if (text.StartsWith(NOT_IN_RANGE))
			{
				Error();
				OutputMsg("Can't seem to find the right container, crafting will stop.");
			}
			break;
		}
		case MSGTYPE_CRAFT_INFO:
		{
			if (actionToDo == CRAFTING_READRECIPE || actionToDo == CRAFTING_SCAN_GROUND_ITEM || actionToDo == CRAFTING_SELECT_CONTAINERS || actionToDo == CRAFTING_USE_COMBINE)
			{
				psMsgCraftingInfo incoming(message);
				csString itemDescription(incoming.craftInfo);

				rcp.ParseRecipe(itemDescription, useMax);
				worldHandler::CloseOpenWindow(csString("Read Book"));
			}
			else if (actionToDo == CRAFTING_SCAN_TABLE)
			{
				// don't parse recipes in storage table, just close the window
				worldHandler::CloseOpenWindow(csString("Read Book"));
			}
			break;
		}
		case MSGTYPE_GUISTORAGE:
		{
			if (!storageInProgress) break;

			psGUIStorageMessage incoming( message );
			PawsManager::GetSingleton().FindWidget("StorageWindow", false)->Hide();

			if (incoming.command == psGUIStorageMessage::ITEMS)
			{
				const char* data = incoming.commandData;
				csRef<iDocumentSystem> xml =  csQueryRegistry<iDocumentSystem> (PawsManager::GetSingleton().GetObjectRegistry());

				csRef<iDocument> itemList  = xml->CreateDocument();

				const char* error = itemList->Parse( data );
				if ( error )
				{
					Error3("Error in XML: %sin %s", error, data );
					return;
				}

				csRef<iDocumentNode> root = itemList->GetRoot();
				if(!root)
				{
					Error2("No XML root in %s", data);
					return;
				}
				csRef<iDocumentNode> topNode = root->GetNode("L");
				if (!topNode)
				{
					Error2("No <L> tag in %s", data);
					return;
				}
				csRef<iDocumentNodeIterator> iter = topNode->GetNodes();
					
				while ( iter->HasNext() )
				{
					csRef<iDocumentNode> item = iter->Next();
					csString name(item->GetAttributeValue("NAME"));
					int count = item->GetAttributeValueAsInt("COUNT");
					if (NewResourceMatch(name, count))
					{
						char * itemID = (char *) item->GetAttributeValue("ID");
						char commandData[200];
						sprintf(commandData, "<T ID=\"%d\" ITEM=\"%s\" COUNT=\"%d\" ITEM_ID=\"%s\" />", 
							worldHandler::GetTargetEID().Unbox(), EscpXML(name).GetData(), count, itemID);
						psGUIStorageMessage outgoing(psGUIStorageMessage::STORE, commandData);
						outgoing.SendMessage();
					}
					
				}

			}
			break;
		}
		case MSGTYPE_VIEW_ITEM:
		{
			psViewItemDescription mesg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
			csString itemName(mesg.itemName);
			if (!categoryMap.Contains(itemName))
			{
				csString desc(mesg.itemDescription);
				desc = desc.Collapse();
				size_t start = desc.Find("Category: ") + 10;
				size_t length = desc.Find("Weight: ", start) - start;
				csString category = desc.Slice(start, length).Trim();
				categoryMap.PutUnique(itemName, category);
			}
			if (awaitingScan) // ignore this non-container item and proceed with scan
			{
				if (unscannedItems.IsEmpty())
				{
					awaitingScan = false;
				}
				else
				{
					EID groundItem = unscannedItems.Pop();
					psViewItemDescription out(groundItem.Unbox(), -1);
					out.SendMessage();
				}
			}
			else if (!tableScanSlots.IsEmpty())
			{
				csString desc(mesg.itemDescription);
				desc = desc.Collapse();
				int slotID = tableScanSlots.Get(0);
				tableScanSlots.DeleteIndex(0);
				size_t start = desc.Find("Quality: ") + 9;
				size_t length = desc.Find(" ", start) - start;
				csString qualityStr = desc.Slice(start, length).Trim();
				int quality = atoi(qualityStr.GetData());
				tableContents[slotID].quality = quality;
				if (tableScanSlots.IsEmpty())
				{
					tableScanInProgress = false;
				}
				else
				{
					psViewItemDescription out(storageContainer.Unbox(), tableScanSlots.Get(0));
					out.SendMessage();
				}
			}
			break;
		}
		case MSGTYPE_GUIINVENTORY:
		{
			psGUIInventoryMessage incoming(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());

			// drop inventory list, if its version is older than the current.
			// this may happen due to UDP latency.
			if (craftInvCache->GetInventoryVersion() >= incoming.version)
				break;
    
			// Set new version in any case (LIST or UPDATE_LIST)
			craftInvCache->SetInventoryVersion(incoming.version);

			// merge the received inventory list into the client's inventory cache
			if (incoming.command == psGUIInventoryMessage::LIST)   // for the whole lot
			{
				craftInvCache->EmptyInventory();   // empty the inventory first
				craftInvCache->SetCacheStatus(psCache::VALID);
			}

			int tableCount = 0;

			for ( size_t z = 0; z < incoming.totalItems; z++ )  // cache inventory items
			{
				craftInvCache->SetInventoryItem(incoming.items[z].slot,
												incoming.items[z].container,
												incoming.items[z].name,
												incoming.items[z].meshName,
												incoming.items[z].materialName,
												incoming.items[z].weight,
												incoming.items[z].size,
												incoming.items[z].stackcount,
												incoming.items[z].itemQuality,
												incoming.items[z].iconImage,
												incoming.items[z].purifyStatus);
			}

			if (incoming.command == psGUIInventoryMessage::UPDATE_LIST)
			{
				for ( size_t z = 0; z < incoming.totalEmptiedSlots; z++)
				{
					craftInvCache->EmptyInventoryItem( incoming.items[incoming.totalItems+z].slot,
														incoming.items[incoming.totalItems+z].container);
				}
			}

			if (watchingInvForNewResources)
			{
				worldHandler::Target(trainerName);
				newResourceSlots.DeleteAll();
				for (size_t i = 0; i < invSnapshot.GetSize(); i++)
				{
					psInventoryCache::CachedItemDescription* cacheSlot = craftInvCache->GetInventoryItem(i+16);
					slot tempSlot;
					int currentSlotCount = cacheSlot == NULL ? 0 : cacheSlot->stackCount;
					int difference = currentSlotCount - invSnapshot[i];
					if (difference > 0)
					{
						tempSlot = slot(cacheSlot->slot, cacheSlot->name, cacheSlot->stackCount, cacheSlot->itemQuality, true);
						tempSlot.container = cacheSlot->containerID;
						csString itemName = cacheSlot->name;
						OutputMsg(csString("    ").Append(difference).Append(" x ").Append(itemName).Append(" q").Append(cacheSlot->itemQuality),TEXT_COLOUR_WHITE);
						newResourceSlots.InsertSorted(tempSlot.slotNo);
						if (!categoryMap.Contains(itemName))
						{
							newResourceNeedsCategory = true;
							psViewItemDescription out(tempSlot.container, tempSlot.slotNo);
							out.SendMessage();
						}
					}
				}
				watchingInvForNewResources = false;
			}
			break;
		}
		case MSGTYPE_VIEW_CONTAINER:
		{
			psViewContainerDescription mesg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
			if (awaitingScan)
			{
				csString desc(mesg.itemDescription);
				size_t index = desc.Find(guardString.GetData());
				if (index > 2) // this is a container guarded by player
				{
					if (containerEID.Unbox() == -1001 && rcp.GetContainerName().CompareNoCase(mesg.itemName) && mesg.contents.GetSize() == 0)
					{
						containerEID = EID(mesg.containerID);
					}
					else if (storageContainer.Unbox() == -1001 && CRAFT_TABLE.CompareNoCase(mesg.itemName))
					{
						storageContainer = EID(mesg.containerID);
					}
				}
				if (unscannedItems.IsEmpty())
				{
					awaitingScan = false;
				}
				else
				{
					EID groundItem = unscannedItems.Pop();
					psViewItemDescription out(groundItem.Unbox(), -1);
					out.SendMessage();
				}
			}
			else if (tableScanInProgress)
			{
				for (int i = 0; i < 16; i++)
				{
					tableContents[i].Clear(i);
				}
				OutputMsg(csString("    scanning table"));
				if ( mesg.hasContents && mesg.contents.GetSize() > 0)
				{
					tableScanSlots.DeleteAll();
					for (size_t i=0; i < mesg.contents.GetSize(); i++)
					{
						int slotID = mesg.contents[i].slotID;
						tableContents[slotID].name = mesg.contents[i].name;
						tableContents[slotID].container = mesg.containerID;
						tableContents[slotID].count = mesg.contents[i].stackCount;
						tableScanSlots.InsertSorted(slotID);
					}
					psViewItemDescription out(mesg.containerID, tableScanSlots.Get(0));
					out.SendMessage();
				}
				else
				{
					OutputMsg(csString("    { all slots empty }"));
					tableScanInProgress = false;
				}
			}
			break;
		}
	}
	message->Reset();
}
void CraftActivity::ScheduleNewPerform(int delay)
{
	csRef<iEventTimer> timer = worldHandler::GetStandardTimer();
	timer->AddTimerEvent(this, delay);
}
bool CraftActivity::Perform(iTimerEvent *ev)
{

	if (!IsOngoing() && !IsFinished())
		return false;

	switch (actionToDo)
	{
		case CRAFTING_INIT:
		{
			DoInit();
			NextAction(MIN_DELAY);
			break;
		}
		case CRAFTING_READRECIPE:
		{
			DoReadRecipe();
			NextAction(MIN_DELAY);
			break;
		}
		case CRAFTING_SCAN_GROUND_ITEM:
		{
			DoScanItem();
			NextAction(MIN_DELAY);
			break;
		}
		case CRAFTING_SELECT_CONTAINERS:
		{
			DoSelectContainers();
			NextAction(MIN_DELAY);
			break;
		}
		case CRAFTING_SCAN_TABLE:
		{
			DoScanTable();
			NextAction(MIN_DELAY);
			break;
		}
		case CRAFTING_USE_COMBINE:
		{
			DoCombine();
			NextAction(craftDelay);
			break;
		}
		case CRAFTING_TAKE_RESULT:
		{
			if (craftFinished)
			{
				DoTakeResult();
				NextAction(MIN_DELAY);
			}
			else if (csGetTicks() - startTime < timeout)
			{
				OutputMsg(csString("Item not yet ready and not yet in timeout, will wait another round."));
				return true;
			}
			else
			{
				OutputMsg(csString("Timed out, will empty container and try again."));
				timesDone = timesDone > 1 ? timesDone-- : 1;
				actionToDo = CRAFTING_INIT;
				worldHandler::TakeAllFromTarget();
				NextAction(MIN_DELAY);
			}
			break;
		}
		case CRAFTING_MAP_CATEGORIES:
		{
			DoMapCategories();
			NextAction(2*MIN_DELAY);
			break;
		}
		case CRAFTING_STORE_RESOURCES:
		{
			DoStoreResources();
			NextAction(2*MIN_DELAY);
			break;
		}
		case CRAFTING_CLOSE_STORAGE:
		{
			DoCloseStorage();
			NextAction(MIN_DELAY);
			break;
		}
		case CRAFTING_TRAIN:
		{
			DoTrain();
			NextAction(2*MIN_DELAY);
			break;
		}
		case CRAFTING_NONE:
		{
			OutputMsg(csString("Craft activity finished."));
		}
	}
	return false;
}

void CraftActivity::NextAction(int delay)
{
	switch (actionToDo)
	{
	case CRAFTING_INIT:
	{
		if (timesDone == 0)
			actionToDo = CRAFTING_READRECIPE;
		else if (timesDone >= timesToRepeat)
			actionToDo = CRAFTING_NONE;
		else
			actionToDo = CRAFTING_USE_COMBINE;
		break;
	}
	case CRAFTING_READRECIPE:
	{
		actionToDo = CRAFTING_SCAN_GROUND_ITEM;
		break;
	}
	case CRAFTING_SCAN_GROUND_ITEM:
	{
		if (unscannedItems.IsEmpty())
			actionToDo = CRAFTING_SELECT_CONTAINERS;
		else actionToDo = CRAFTING_SCAN_GROUND_ITEM;
		break;
	}
	case CRAFTING_SELECT_CONTAINERS:
	{
		actionToDo = CRAFTING_SCAN_TABLE;
		break;
	}
	case CRAFTING_SCAN_TABLE:
	{
		if (tableScanInProgress)
			actionToDo = CRAFTING_SCAN_TABLE;
		else
			actionToDo = CRAFTING_USE_COMBINE;
		break;
	}
	case CRAFTING_USE_COMBINE:
	{
		actionToDo = CRAFTING_TAKE_RESULT;
		break;
	}
	case CRAFTING_TAKE_RESULT:
	{
		if (watchingInvForNewResources)
			actionToDo = CRAFTING_TAKE_RESULT;
		else
			actionToDo = CRAFTING_MAP_CATEGORIES;
		break;
	}
	case CRAFTING_MAP_CATEGORIES:
	{
		if (newResourceNeedsCategory)
			actionToDo = CRAFTING_MAP_CATEGORIES;
		else
			actionToDo = CRAFTING_STORE_RESOURCES;
		break;
	}
	case CRAFTING_STORE_RESOURCES:
	{
		actionToDo = CRAFTING_CLOSE_STORAGE;
		break;
	}
	case CRAFTING_CLOSE_STORAGE:
	{
		if (storageInProgress)
			actionToDo = CRAFTING_CLOSE_STORAGE;
		else
			actionToDo = CRAFTING_TRAIN;
		break;
	}
	case CRAFTING_TRAIN:
	{
		actionToDo = CRAFTING_INIT;
		break;
	}
	}
	ScheduleNewPerform(delay);
}

void CraftActivity::DoInit()
{
	actionToDo = CRAFTING_INIT;
	needToTrain = false;
	craftFinished = false;
	awaitingScan = false;
	tableScanInProgress = false;
	storageInProgress = false;
	lootOnGround = false;
	newResourceNeedsCategory = false;
	preResultInvVersion = 0;
	logText = csString("");
	startTime = csGetTicks();

	if (timesDone == 0 && cmd != NULL && cmd.Length() > 2)
	{
		WordArray words(cmd,false);

		if (words.GetCount() > 0)
			craftDelay = words.GetInt(0);
		else craftDelay = MAX_DELAY;

		if (words.GetCount() > 1 && words[1] == "m")
			  useMax = false;
		else useMax = true;

		if (words.GetCount() > 2)
			timeout = words.GetInt(2);
		else timeout = craftDelay;

		csString recipeName = worldHandler::GetBrainItemName();
		if (recipeName.CompareNoCase(""))
		{
			Error();
			actionToDo = CRAFTING_NONE;
			OutputMsg(csString("No blueprint equipped, please equip the bp for what you want to craft and /bot craft again."));
			return;
		}
		else
			rcp.setRecipeName(recipeName);
		
		containerEID = EID(-1001);
		storageContainer = EID(-1001);
		unscannedItems = worldHandler::FindItemsInReach();
		guardString = csString("Guarded by: ").Append(psengine->GetCelClient()->GetMainPlayer()->GetName());
	}

	if (useMax)
		logText = rcp.GetItemName()+LOG_TEXT_SEP+"Max"+LOG_TEXT_SEP;
	else logText = rcp.GetItemName()+LOG_TEXT_SEP+"min"+LOG_TEXT_SEP;
}

void CraftActivity::DoReadRecipe()
{
	psViewItemDescription out(CONTAINER_INVENTORY_EQUIPMENT, PSCHARACTER_SLOT_MIND);
	out.SendMessage();
	OutputMsg(csString("    reading the recipe"));
}
void CraftActivity::DoScanItem()
{
	if (!unscannedItems.IsEmpty())
	{
		EID groundItem = unscannedItems.Pop();
		psViewItemDescription out(groundItem.Unbox(), -1);
		out.SendMessage();
		awaitingScan = true;
	}
}
void CraftActivity::DoSelectContainers()
{
	if (containerEID.Unbox() == -1001)
	{
		OutputMsg(csString("An empty crafting container not found: ").Append(rcp.GetContainerName()).Append(" bot stopping"));
		Error();
		return;
	}
	if (storageContainer.Unbox() == -1001)
	{
		OutputMsg(csString("Storage Craft-Table not found, bot Stopping"));
		Error();
		return;
	}
	OutputMsg(csString("    crafting container ").Append(rcp.GetContainerName()).Append(": ").Append(containerEID.Unbox()));
	OutputMsg(csString("    storage table: ").Append(storageContainer.Unbox()));
	tableScanInProgress = true;
	psViewItemDescription out(storageContainer.Unbox(), -1);
	out.SendMessage();
}
void CraftActivity::DoScanTable()
{

}
void CraftActivity::DoCombine()
{
	char out[200];
	sprintf(out, "Done %d items, %d left to do.", timesDone, timesToRepeat-timesDone);
	OutputMsg(csString(out));

	if (!worldHandler::IsInBrain(rcp.GetRecipeName()))
	{
		int bpSlot = FindTableSlot(rcp.GetRecipeName());
		if (bpSlot >= 0)
		{
			int remainingRounds = timesToRepeat-timesDone;
			int moveQuantity = remainingRounds > tableContents[bpSlot].count ? tableContents[bpSlot].count : remainingRounds;
			moveQuantity = moveQuantity > 100 ? 100 : moveQuantity;
			psSlotMovementMsg msg(tableContents[bpSlot].container, bpSlot, CONTAINER_INVENTORY_EQUIPMENT, PSCHARACTER_SLOT_MIND, moveQuantity);
			msg.SendMessage();
			tableContents[bpSlot].count -= moveQuantity;
			OutputMsg(csString("    equipped ").Append(moveQuantity).Append(" more blueprints"));
		}
		else
		{
			OutputMsg(csString("Cannot find anymore blueprints"));
			Error();
			return;
		}
	}


	if (!worldHandler::TargetEID(containerEID))
	{
		OutputMsg(csString("Could not find container within reach!"));
		Error();
		return;
	}

	csString contName = worldHandler::GetTargetName().Collapse();
	if (!contName.CompareNoCase(rcp.GetContainerName()))
	{
		OutputMsg(rcp.GetContainerName());
		OutputMsg(contName);
		OutputMsg(csString("Wrong container targeted! Kindly target the right container for the task and /bot craft again."));
		Error();
		return;
	}

	if (!worldHandler::OpenTarget())
	{
		OutputMsg(csString("Could not open the container!"));
		Error();
		return;
	}
	OutputMsg(csString("Moving ingredients"));

	int bundleSlot = FindTableSlot(rcp.GetBundleName());
	if (bundleSlot >= 0)
	{
		psSlotMovementMsg msg(tableContents[bundleSlot].container, bundleSlot, containerEID.Unbox(), 0, 1);
		msg.SendMessage();
		tableContents[bundleSlot].count--;
		OutputMsg(csString("    ").Append(rcp.GetBundleName()).Append(" x1"));
		OutputMsg(csString("Starting combine."));
		worldHandler::UseTarget();
		logText = logText + rcp.GetBundleName();
		return;
	}
	csHash<int, csString>::GlobalIterator iterIngredients(rcp.GetIngredientsList()->GetIterator());

	int nextEmptySlot = 0;
	if (!iterIngredients.HasNext())
	{
		OutputMsg(csString("    Empty ingredients list!"));
		Error();
		return;
	}
	while (iterIngredients.HasNext())
	{
		csString itemName;
		int quantity = iterIngredients.Next(itemName);

		char out[1000];
		sprintf(out, "Ingredient %s: %d", itemName.GetData(), quantity);

		int slotNo = FindTableSlot(itemName, quantity);
		if (slotNo < 0)
		{
			OutputMsg(csString("        not enough ").Append(itemName).Append(", bot stopping."));
			Error();
			return;
		}
		psSlotMovementMsg msg(tableContents[slotNo].container, slotNo, containerEID.Unbox(), nextEmptySlot, quantity);
		msg.SendMessage();
		nextEmptySlot += 1;
		tableContents[slotNo].count -= quantity;
		logText = logText+csString(out)+ csString(" ");
		OutputMsg(csString("    ").Append(itemName).Append(" x").Append(quantity));
	}
	OutputMsg(csString("Starting combine."));
	worldHandler::CombineContentsInTarget();
}
void CraftActivity::DoTakeResult()
{
	if (lootOnGround)
	{
		OutputMsg(csString("Loot on ground, bot stopping."));
		Error();
		return;
	}

	if (watchingInvForNewResources)
	{
		OutputMsg(csString("    waiting for /takeall..."));
		return;
	}

	if (preResultInvVersion != 0) return;
	OutputMsg(csString("Taking output..."));
	preResultInvVersion = craftInvCache->GetInventoryVersion();
	invSnapshot = InventorySnapshot();
	worldHandler::TakeAllFromTarget();
	watchingInvForNewResources = true;
}

csArray<int> CraftActivity::InventorySnapshot()
{
	csArray<int> result(80);
	for (size_t i = 0; i < 80; i++)
	{
		psInventoryCache::CachedItemDescription* slot = craftInvCache->GetInventoryItem(i+16);
		result.Put(i, slot == NULL ? 0 : slot->stackCount);
	}
	return result;
}

void CraftActivity::DoMapCategories()
{
	if (craftInvCache->GetInventoryVersion() == preResultInvVersion)
	{
		OutputMsg(csString("    waiting for /takeall..."));
		return;
	}
	newResourceNeedsCategory = false;
	for (size_t i = 0; i < newResourceSlots.GetSize(); i++)
	{
		csString itemName = craftInvCache->GetInventoryItem(newResourceSlots.Get(i))->name;
		if (!categoryMap.Contains(itemName))
		{
			newResourceNeedsCategory = true;
			psViewItemDescription out(-2, newResourceSlots.Get(i));
			out.SendMessage();

			OutputMsg(itemName.Append(" has no category."));
		}
	}
	if (!newResourceNeedsCategory)
	{ // initiate storage session
		storageInProgress = true;
		psGUIStorageMessage exchange(psGUIStorageMessage::REQUEST,csString("<R TYPE=\"STORE\"/>"));
		exchange.SendMessage();
	}
}

void CraftActivity::DoStoreResources()
{
	csSet<csString> uniqueCategories;
	for (size_t i = 0; i < newResourceSlots.GetSize(); i++)
	{
		csString itemName = craftInvCache->GetInventoryItem(newResourceSlots.Get(i))->name;
		uniqueCategories.Add(categoryMap.Get(itemName,csString("")));
	}

	csSet<csString>::GlobalIterator categories = uniqueCategories.GetIterator();
    while (categories.HasNext())
    {
		csString category = categories.Next();
		csString commandData;
		commandData.Format("<C ID=\"%d\" CATEGORY=\"%s\" />", worldHandler::GetTargetEID().Unbox(),
							EscpXML(category).GetData());

		psGUIStorageMessage outgoing(psGUIStorageMessage::CATEGORY, commandData);
		outgoing.SendMessage();
	}
}

void CraftActivity::DoCloseStorage()
{
	//csArray<int> slotsCopy = csArray<int>(newResourceSlots);
	//for (size_t i = 0; i < slotsCopy.GetSize(); i++)
	//{
	//	int64_t stackCount = craftInvCache->GetInventoryItem(slotsCopy.Get(i))->stackCount;
	//	if (stackCount == 0)
	//	{
	//		printf("before delete\n");
	//		newResourceSlots.Delete(slotsCopy.Get(i));
	//		printf("after delete\n");
	//	}
	//}
	//if (newResourceSlots.IsEmpty())
	{
		storageInProgress = false;
		csString commandData;
		commandData.Format("<C ID=\"%d\"/>", worldHandler::GetTargetEID().Unbox());
		psGUIStorageMessage outgoing(psGUIStorageMessage::CANCEL, commandData);
		outgoing.SendMessage();
	}
	//else
	//{
	//	OutputMsg("storage not yet complete");
	//}
}
void CraftActivity::DoTrain()
{
	if (needToTrain)
	{
		OutputMsg(csString("Training ").Append(skillToTrain));
		if (!worldHandler::TrainSkillMax(skillToTrain))
			OutputMsg(csString("    trainer not in range"));
		else
			needToTrain = false;
	}
	Finish();
}
int CraftActivity::FindTableSlot(csString name, int count)
{
	for (int i = 0; i < 16; i++)
	{
		if (tableContents[i].count > 0)
		{
			if(tableContents[i].name.Compare(name) && tableContents[i].count >= count)
			{
				return i;
			}
		}
	}
	return -1;
}

bool CraftActivity::NewResourceMatch(csString name, int count)
{
	bool result = false;
	csArray<int> emptySlots;
	for (size_t i = 0; i < newResourceSlots.GetSize(); i++)
	{
		psInventoryCache::CachedItemDescription* cacheSlot = craftInvCache->GetInventoryItem(newResourceSlots.Get(i));
		if (cacheSlot == NULL) // this slot was bound for storage and is now empty
		{
			emptySlots.Push(newResourceSlots.Get(i));
			continue;
		}
		result |= (name.Compare(cacheSlot->name) && count == cacheSlot->stackCount);
	}

	// remove newly empty slots from resource array
	for (size_t i = 0; i < emptySlots.GetSize(); i++)
	{
		newResourceSlots.Delete(emptySlots.Get(i));
	}
	return result;
}

/*************************Repair Activity*****************/

RepairActivity::RepairActivity()
{
	name = csString("repair");
	textcolour = TEXT_COLOUR;
	needToTrain = false;
	craftDelay = 5000;
	repairFinished = false;
}
RepairActivity::~RepairActivity()
{

}
void RepairActivity::HandleMessage(MsgEntry* message )
{
	if (!IsOngoing())
		return;

	switch (message->GetType() )
		{
		case MSGTYPE_SYSTEM:
			{
				psSystemMessage incoming(message);
				csString text = incoming.msgline;

				if (text.StartsWith(EXPLORE_TOOL_WORN))
				{
					OutputMsg(csString("Tool worn-out, will change it."));
					//get a new tool
					csString toolName = worldHandler::GetEquippedToolName();

					if (toolSlot == NULL)
						toolSlot = worldHandler::GetNextFreeInvSlot();

					if (toolSlot == NULL)
					{
						Error();
						OutputMsg(csString("Full inventory and no place to move worn-out tool, bot will stop!"));
						break;
					}

					worldHandler::MoveItems(CONTAINER_INVENTORY_EQUIPMENT, PSCHARACTER_SLOT_RIGHTHAND, toolSlot->ContainerID(), toolSlot->ID(), 1);
					worldHandler::Equip(toolName);
				}
				else if (text.StartsWith(FINISHED_WORKING))
					repairFinished = true;
			}
		}
		

}


void RepairActivity::ScheduleNewPerform(int delay)
{
	csRef<iEventTimer> timer = worldHandler::GetStandardTimer();
	timer->AddTimerEvent(this, delay);
}
bool RepairActivity::Perform(iTimerEvent *ev)
{
	if (!IsOngoing() && !IsFinished())
		return false;

	switch (actionToDo)
	{
		case REPAIR_INIT:
		{
			DoInit();
			NextAction(MIN_DELAY);
			break;
		}
		case REPAIR_WORK:
		{
			DoRepairWork();
			NextAction(MAX_DELAY);
			break;
		}
		case REPAIR_TRAIN:
		{
			if (repairFinished)
			{
				DoTrain();
				NextAction(2*MIN_DELAY);
			}
			else ScheduleNewPerform(2*MIN_DELAY);
			break;
		}
		case REPAIR_NONE:
		{
			Finish();
			break;
		}
	}
	return false;
}

void RepairActivity::StartWork()
{
	OutputMsg(csString("Repair activity started."));
	ScheduleNewPerform(MIN_DELAY);
}
void RepairActivity::WrapUp()
{
	if (logger == NULL)
		OutputMsg(csString("No logger available, so activities will NOT be logged."));
	else
		logger->Write(CSV_CRAFTING, logText);
}
void RepairActivity::DoInit()
{
	needToTrain = false;
	craftDelay = MAX_DELAY;
	repairFinished = false;

	logText = csString("");
	startTime = csGetTicks();
	actionToDo = REPAIR_NONE;

	if (timesDone == 0 && cmd != NULL && cmd.Length() > 1)
	{
		toolSlot = NULL;
		WordArray words(cmd,false);

		if (words.GetCount() > 0)
			timeout = words.GetInt(0);
		else timeout = 100*MAX_DELAY;

		containerEID = worldHandler::GetTargetEID();

		if (containerEID == -1)
		{
			Error();
			actionToDo = REPAIR_NONE;
			OutputMsg(csString("Kindly target (open) the container you want repaired and then /bot repair again."));
			return;
		}
	}
	actionToDo = REPAIR_WORK;
}

void RepairActivity::NextAction(int delay)
{
	switch (actionToDo)
	{
	case REPAIR_INIT:
	{
		actionToDo = REPAIR_WORK;
		break;
	}
	case REPAIR_WORK:
	{
		actionToDo = REPAIR_TRAIN;
		break;
	}
	case REPAIR_TRAIN:
	{
		if (timesDone < timesToRepeat)
			actionToDo = REPAIR_INIT;
		else actionToDo = REPAIR_NONE;
		break;
	}
	}
	ScheduleNewPerform(delay);
}

void RepairActivity::DoRepairWork()
{
	OutputMsg(csString("Starting repair work."));
	//do the actual repair work
	worldHandler::TargetEID(containerEID);
	//worldHandler::OpenTargetEID(containerEID);
	worldHandler::ExecCmd(csString("/repair"));
}

void RepairActivity::DoTrain()
{
	OutputMsg(csString("Repair finished."));
	Finish();
	if (needToTrain)
	{
		if (!worldHandler::TrainSkillMax(skillToTrain))
			OutputMsg(csString("Trainer not in range, so can't train."));
		needToTrain = false;
	}
}

/************************* Bundle Activity *************************/

BundleActivity::BundleActivity()
{
	name = csString("bundle");
	textcolour = TEXT_COLOUR;
}
BundleActivity::~BundleActivity()
{
}
void BundleActivity::StartWork()
{
	OutputMsg(csString("Bundle activity started."));
	ScheduleNewPerform(MIN_DELAY);
}
void BundleActivity::WrapUp()
{
}

void BundleActivity::HandleMessage(MsgEntry* message )
{
	if (!IsOngoing())
		return;

	switch (message->GetType() )
	{
		case MSGTYPE_SYSTEM:
		{
			psSystemMessage incoming(message);
			csString text = incoming.msgline;

			if (text.StartsWith(CONTAINER_WORN))
			{
				Error();
				OutputMsg("Container worn-out, bundling will stop.");
			}
			else if (text.StartsWith(TOOL_WORN))
			{
				Error();
				OutputMsg("Tool worn-out, bundling will stop.");
			}
			else if (text.StartsWith(NOT_IN_RANGE))
			{
				Error();
				OutputMsg("Can't seem to reach the right container, bundling will stop.");
			}
			break;
		}
		case MSGTYPE_CRAFT_INFO:
		{
			if (actionToDo == BUNDLE_READRECIPE || actionToDo == BUNDLE_SCAN_GROUND_ITEM || actionToDo == BUNDLE_SELECT_CONTAINERS || actionToDo == BUNDLE_USE_COMBINE)
			{
				psMsgCraftingInfo incoming(message);
				csString itemDescription(incoming.craftInfo);

				if (inferQuantities)
				{
					rcp.ParseRecipe(itemDescription, true);
					extraRcp.ParseRecipe(itemDescription, false);
				}
				else
				{
					rcp.ParseRecipe(itemDescription, useMax);
				}
			}
			worldHandler::CloseOpenWindow(csString("Read Book"));
			break;
		}
		case MSGTYPE_VIEW_ITEM:
		{
			psViewItemDescription mesg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
			csString itemName(mesg.itemName);
			if (awaitingScan)
			{
				OutputMsg(csString("    ignoring non-container on ground: ").Append(itemName));
				if (unscannedItems.IsEmpty())
				{
					awaitingScan = false;
				}
				else
				{
					EID groundItem = unscannedItems.Pop();
					psViewItemDescription out(groundItem.Unbox(), -1);
					out.SendMessage();
				}
			}
			else if (!tableScanSlots.IsEmpty())
			{
				csString desc(mesg.itemDescription);
				desc = desc.Collapse();
				int slotID = tableScanSlots.Get(0);
				tableScanSlots.DeleteIndex(0);
				size_t start = desc.Find("Quality: ") + 9;
				size_t length = desc.Find(" ", start) - start;
				csString qualityStr = desc.Slice(start, length).Trim();
				int quality = atoi(qualityStr.GetData());
				tableContents[slotID].quality = quality;
				if (tableScanSlots.IsEmpty())
				{
					tableScanInProgress = false;
				}
				else
				{
					psViewItemDescription out(storageContainer.Unbox(), tableScanSlots.Get(0));
					out.SendMessage();
				}
			}
			break;
		}
		case MSGTYPE_VIEW_CONTAINER:
		{
			psViewContainerDescription mesg(message, psengine->GetNetManager()->GetConnection()->GetAccessPointers());
			if (awaitingScan)
			{
				csString desc(mesg.itemDescription);
				csString longName = csString(mesg.itemName).Append(" EID: ").Append(mesg.containerID);
				size_t index = desc.Find(guardString.GetData());
				if (index > 2) // this is a container guarded by player
				{
					if (containerEID.Unbox() == -1001 && !rcp.GetContainerName().CompareNoCase(mesg.itemName) && mesg.contents.GetSize() == 0)
					{
						OutputMsg(csString("Work container: ").Append(longName));
						containerEID = EID(mesg.containerID);
					}
					else if (storageContainer.Unbox() == -1001 && CRAFT_TABLE.CompareNoCase(mesg.itemName))
					{
						OutputMsg(csString("Storage container: ").Append(longName));
						storageContainer = EID(mesg.containerID);
					}
					else
					{
						OutputMsg(csString("    container not selected: ").Append(longName));
					}
				}
				else
				{
					OutputMsg(csString("    container not selected because not guarded by you: ").Append(longName));
				}
				if (unscannedItems.IsEmpty())
				{
					awaitingScan = false;
				}
				else
				{
					EID groundItem = unscannedItems.Pop();
					psViewItemDescription out(groundItem.Unbox(), -1);
					out.SendMessage();
				}
			}
			else if (tableScanInProgress)
			{
				for (int i = 0; i < 16; i++)
				{
					tableContents[i].Clear(i);
				}
				if ( mesg.hasContents && mesg.contents.GetSize() > 0)
				{
					tableScanSlots.DeleteAll();
					for (size_t i=0; i < mesg.contents.GetSize(); i++)
					{
						int slotID = mesg.contents[i].slotID;
						tableContents[slotID].name = mesg.contents[i].name;
						tableContents[slotID].container = mesg.containerID;
						tableContents[slotID].count = mesg.contents[i].stackCount;
						tableScanSlots.InsertSorted(slotID);
					}
					psViewItemDescription out(mesg.containerID, tableScanSlots.Get(0));
					out.SendMessage();
				}
				else
				{
					OutputMsg(csString("    { all slots empty }"));
					tableScanInProgress = false;
				}
			}
			else if (!bundleFinished)
			{
				if ( mesg.hasContents && mesg.contents.GetSize() > 0 && mesg.containerID == containerEID.Unbox())
				{
					tableScanSlots.DeleteAll();
					for (size_t i=0; i < mesg.contents.GetSize(); i++)
					{
						if (mesg.contents[i].slotID == 0 && mesg.contents[i].stackCount == 1)
						{
							bundleFinished = true;
							break;
						}
					}
				}
			}
			break;
		}
	}
	message->Reset();
}
void BundleActivity::ScheduleNewPerform(int delay)
{
	csRef<iEventTimer> timer = worldHandler::GetStandardTimer();
	timer->AddTimerEvent(this, delay);
}
bool BundleActivity::Perform(iTimerEvent *ev)
{

	if (!IsOngoing() && !IsFinished())
		return false;

	switch (actionToDo)
	{
		case BUNDLE_INIT:
		{
			DoInit();
			NextAction(10);
			break;
		}
		case BUNDLE_READRECIPE:
		{
			DoReadRecipe();
			NextAction(MIN_DELAY);
			break;
		}
		case BUNDLE_SCAN_GROUND_ITEM:
		{
			DoScanItem();
			NextAction(MIN_DELAY);
			break;
		}
		case BUNDLE_SELECT_CONTAINERS:
		{
			DoSelectContainers();
			NextAction(MIN_DELAY);
			break;
		}
		case BUNDLE_SCAN_TABLE:
		{
			DoScanTable();
			NextAction(MIN_DELAY);
			break;
		}
		case BUNDLE_INFER_RECIPE:
		{
			DoInferRecipe();
			NextAction(MIN_DELAY);
			break;
		}
		case BUNDLE_REPORT:
		{
			DoReport();
			NextAction(MAX_DELAY*2);
			break;
		}
		case BUNDLE_USE_COMBINE:
		{
			DoCombine();
			NextAction(10);
			break;
		}
		case BUNDLE_MOVE_OUTPUT:
		{
			DoMoveOutput();
			NextAction(10);
			break;
		}
		case BUNDLE_NONE:
		{
			worldHandler::OpenTargetEID(storageContainer);
			OutputMsg(csString("Bundle activity finished."));
		}
	}
	return false;
}

void BundleActivity::NextAction(int delay)
{
	switch (actionToDo)
	{
	case BUNDLE_INIT:
	{
		if (timesDone == 0)
			actionToDo = BUNDLE_READRECIPE;
		else if (timesDone >= timesToRepeat)
			actionToDo = BUNDLE_NONE;
		else
			actionToDo = BUNDLE_USE_COMBINE;
		break;
	}
	case BUNDLE_READRECIPE:
	{
		actionToDo = BUNDLE_SCAN_GROUND_ITEM;
		break;
	}
	case BUNDLE_SCAN_GROUND_ITEM:
	{
		if (unscannedItems.IsEmpty())
			actionToDo = BUNDLE_SELECT_CONTAINERS;
		else actionToDo = BUNDLE_SCAN_GROUND_ITEM;
		break;
	}
	case BUNDLE_SELECT_CONTAINERS:
	{
		actionToDo = BUNDLE_SCAN_TABLE;
		break;
	}
	case BUNDLE_SCAN_TABLE:
	{
		if (tableScanInProgress)
			actionToDo = BUNDLE_SCAN_TABLE;
		else if (inferQuantities)
			actionToDo = BUNDLE_INFER_RECIPE;
		else
			actionToDo = BUNDLE_REPORT;
		break;
	}
	case BUNDLE_INFER_RECIPE:
	{
		actionToDo = BUNDLE_REPORT;
		break;
	}
	case BUNDLE_REPORT:
	{
		actionToDo = BUNDLE_USE_COMBINE;
		break;
	}
	case BUNDLE_USE_COMBINE:
	{
		actionToDo = BUNDLE_MOVE_OUTPUT;
		break;
	}
	case BUNDLE_MOVE_OUTPUT:
	{
		actionToDo = BUNDLE_MOVE_OUTPUT;
		break;
	}
	}
	ScheduleNewPerform(delay);
}

void BundleActivity::DoInit()
{
	actionToDo = BUNDLE_INIT;
	awaitingScan = false;
	bundleFinished = false;
	tableScanInProgress = false;
	startTime = csGetTicks();

	if (timesDone == 0 && cmd != NULL && cmd.Length() > 0)
	{
		WordArray words(cmd,false);
		if (words.GetCount() != 1)
		{
			Error();
			actionToDo = BUNDLE_NONE;
			OutputMsg(csString("Usage: /bot bundle total_times m/M/i"));
			return;
		}

		inferQuantities = words[0] == "i";
		if (words[0] == "m") 
			useMax = false;
		else
			useMax = true;

		csString recipeName = worldHandler::GetBrainItemName();
		if (recipeName.CompareNoCase(""))
		{
			Error();
			actionToDo = BUNDLE_NONE;
			OutputMsg(csString("No blueprint equipped, please equip the bp for what you want to bundle and /bot bundle again."));
			return;
		}
		else
			rcp.setRecipeName(recipeName);
		
		containerEID = EID(-1001);
		storageContainer = EID(-1001);
		unscannedItems = worldHandler::FindItemsInReach();
		guardString = csString("Guarded by: ").Append(psengine->GetCelClient()->GetMainPlayer()->GetName());
	}
}

void BundleActivity::DoReadRecipe()
{
	psViewItemDescription out(CONTAINER_INVENTORY_EQUIPMENT, PSCHARACTER_SLOT_MIND);
	out.SendMessage();
}
void BundleActivity::DoScanItem()
{
	if (!unscannedItems.IsEmpty())
	{
		EID groundItem = unscannedItems.Pop();
		psViewItemDescription out(groundItem.Unbox(), -1);
		out.SendMessage();
		awaitingScan = true;
	}
}
void BundleActivity::DoSelectContainers()
{
	if (containerEID.Unbox() == -1001)
	{
		OutputMsg(csString("An empty crafting container not found, bot stopping"));
		Error();
		return;
	}
	if (storageContainer.Unbox() == -1001)
	{
		OutputMsg(csString("Storage Craft-Table not found, bot Stopping"));
		Error();
		return;
	}
	tableScanInProgress = true;
	psViewItemDescription out(storageContainer.Unbox(), -1);
	out.SendMessage();
}
void BundleActivity::DoScanTable()
{
	bundleSlot = -1;
	for (int i = 0; i < 16; i++)
	{
		if (tableContents[i].count == 0)
		{
			bundleSlot = i;
			break;
		}
	}
	if (bundleSlot == -1)
	{
		OutputMsg(csString("An empty slot in the storage container was not found for placing bundles"));
		Error();
		return;
	}
}
void BundleActivity::DoInferRecipe()
{
	csHash<int, csString>::GlobalIterator iterIngredients(extraRcp.GetIngredientsList()->GetIterator());
	while (iterIngredients.HasNext())
	{
		csString itemName;
		int minQuantity = iterIngredients.Next(itemName);
		int maxQuantity = rcp.GetQuantity(itemName);
		int slot = FindTableSlot(itemName);
		if (slot == -1)
		{
			OutputMsg(csString("Not found in storage table: ").Append(itemName));
			OutputMsg(csString("Bot stopping"), TEXT_COLOUR_WHITE);
			Error();
			return;		
		}
		int64_t total = tableContents[slot].count;
		int desiredQuantity = total / timesToRepeat;
		if (desiredQuantity < minQuantity)
		{
			OutputMsg(csString("The inferred quantity is lower than the min"));
			OutputMsg(csString("    item: ").Append(itemName));
			OutputMsg(csString("    stack count: ").Append(total));
			OutputMsg(csString("    desired bundles: ").Append(timesToRepeat));
			OutputMsg(csString("    inferred quanity: ").Append(desiredQuantity));
			OutputMsg(csString("    min quanity: ").Append(minQuantity));
			OutputMsg(csString("Bot stopping"), TEXT_COLOUR_WHITE);
			Error();
			return;
		}
		if (desiredQuantity < maxQuantity)
		{
			rcp.SetQuantity(itemName, desiredQuantity);
		}
	}
}
void BundleActivity::DoReport()
{
	OutputMsg(csString("\nBundle making is about to begin for:"));
	OutputMsg(csString("    ").Append(rcp.GetBundleName()).Append(" x").Append(timesToRepeat), TEXT_COLOUR_WHITE);
	OutputMsg(csString("Using ingredients:"));
	csHash<int, csString>::GlobalIterator iterIngredients(rcp.GetIngredientsList()->GetIterator());
	if (!iterIngredients.HasNext())
	{
		OutputMsg(csString("    Empty ingredients list!"), TEXT_COLOUR_WHITE);
		Error();
		return;
	}
	while (iterIngredients.HasNext())
	{
		csString itemName;
		int quantity = iterIngredients.Next(itemName);
		OutputMsg(csString("    ").Append(itemName).Append(" x").Append(quantity), TEXT_COLOUR_WHITE);
	}
	OutputMsg(csString("work will commence in 15 seconds..."));
}

void BundleActivity::DoCombine()
{
	char out[200];
	sprintf(out, "Done %d items, %d left to do.", timesDone, timesToRepeat-timesDone);
	OutputMsg(csString(out));

	if (!worldHandler::IsInBrain(rcp.GetRecipeName()))
	{
		OutputMsg(csString("Blueprint not equipped"));
		Error();
		return;
	}

	if (!worldHandler::TargetEID(containerEID))
	{
		OutputMsg(csString("Could not find container within reach!"));
		Error();
		return;
	}

	csString contName = worldHandler::GetTargetName().Collapse();
	if (contName.CompareNoCase(rcp.GetContainerName()))
	{
		OutputMsg(rcp.GetContainerName());
		OutputMsg(contName);
		OutputMsg(csString("Wrong container targeted!"));
		Error();
		return;
	}

	if (!worldHandler::OpenTarget())
	{
		OutputMsg(csString("Could not open the container!"));
		Error();
		return;
	}

	csHash<int, csString>::GlobalIterator iterIngredients(rcp.GetIngredientsList()->GetIterator());

	// start ingredients in slot 1 so that slot 0 will only ever contain a bundle
	int nextEmptySlot = 1;
	if (!iterIngredients.HasNext())
	{
		OutputMsg(csString("    Empty ingredients list!"));
		Error();
		return;
	}
	while (iterIngredients.HasNext())
	{
		csString itemName;
		int quantity = iterIngredients.Next(itemName);
		int slotNo = FindTableSlot(itemName, quantity);
		if (slotNo < 0)
		{
			OutputMsg(csString("Not enough ").Append(itemName).Append(" in storage table, stopping."), TEXT_COLOUR_WHITE);
			Error();
			worldHandler::OpenTargetEID(storageContainer);
			return;
		}
		psSlotMovementMsg msg(tableContents[slotNo].container, slotNo, containerEID.Unbox(), nextEmptySlot, quantity);
		msg.SendMessage();
		nextEmptySlot += 1;
		tableContents[slotNo].count -= quantity;
	}
	worldHandler::CombineContentsInTarget();
}
void BundleActivity::DoMoveOutput()
{
	if (bundleFinished)
	{
		psSlotMovementMsg msg(containerEID.Unbox(), 0, storageContainer.Unbox(), bundleSlot, 1);
		msg.SendMessage();
		Finish();
		actionToDo = BUNDLE_INIT;
	}
}
int BundleActivity::FindTableSlot(csString name, int count)
{
	for (int i = 0; i < 16; i++)
	{
		if (tableContents[i].count > 0)
		{
			if(tableContents[i].name.Compare(name) && tableContents[i].count >= count)
			{
				return i;
			}
		}
	}
	return -1;
}