/*
 * worldhandler.h
 *
 *  Created on: 22 Jul 2015
 *      Author: Diana Coman
 */

#ifndef WORLDHANDLER_H_
#define WORLDHANDLER_H_

#include <csutil/timer.h>
#include "net/cmdhandler.h"
#include "gui/pawsslot.h"
#include "pscelclient.h"
#include "psinventorycache.h"

const csString trainerName = csString("Heina Draggenfort");

static const csString ITEM_NAME_FOR_TINY_CLAIMS = csString("Little Bit O' Nothing");
static const csString ITEM_NAME_FOR_SMALL_CLAIMS = csString("Coarse Frangible Thread");

static const char* MVM_NONE = "none";
static const char* MVM_FORWARD = "forward";
static const char* MVM_BACKWARD = "backward";
static const char* MVM_ROT_LEFT = "rotate left";
static const char* MVM_ROT_RIGHT = "rotate right";
static const char* MVM_STRAFE_LEFT = "strafe left";
static const char* MVM_STRAFE_RIGHT = "strafe right";
static const char* MVM_JUMP = "jump";
static const char* MVM_WALK = "normal";

static const float WALK_DELTA = 3.0f;
static const float GUARD_DISTANCE = 10.0f;
static const float PICKUP_DISTANCE = 5.0f;

//handlers for interaction with the game world and original client core
class worldHandler
{
	//auxiliary methods for item handling
public:

	//wrapper for any command to pass directly
	static void ExecCmd(csString cmd);

	//movement
	static void StepToward(const csVector3 &pos);
	static void SendDRUpdate();
	static void StepForward();
	static void StepBackward();
	static void StopAllMovement();
	static void TurnRight(float rad);
	static void TurnToPoint(float x, float z);
	static void Turn(float angle);
	static void SetRotation(float yRot);
	static csVector3 GetCurrentPos();
	static float GetCurrentRot();
	static void Move(const char* moveName, float distance = 1.0f);

	//timer (crystalspace)
	static csPtr<iEventTimer> GetStandardTimer();//iObjectRegistry* object_reg);

	//tests and finder methods
	static bool IsInBrain(csString itemName);
	static bool IsWithinReach(csString name);
    static pawsSlot* FindInventoryItemSlot(csString itemName);
    static psInventoryCache::CachedItemDescription* FindItemSlot(csString itemName, bool anywhere);


    //item action methods
    static bool MoveItems(int fromContainer, int fromSlot, int toContainer, int toSlot, int64_t quantity);
    static bool Equip(csString itemName);
    static bool Dequip(csString itemName);
    static bool EquipRecipe(csString itemName);
    static csString GetEquippedToolName();
    static int GetEquippedToolQuality();
    static csString GetBrainItemName();
    static bool PickupEID(EID eid);
    static bool DropFromSlot(int fromSlot, int quantity);

    //containers and world objects targeting/using/combining/opening/properties
    static bool Target(csString containerName);
    static bool TargetEID(EID eid);
    static csString GetTargetName();
    static bool OpenTarget();
    static bool OpenTargetEID(EID eid);
	static void RemoveObject(EID eid);
	static int Nuke(int dist);
	static int RequestServerWorld();
    static bool CombineContentsInTarget();
    static bool UseTarget();
    static bool PicklockEID(EID eid);
    static bool TakeAllFromTarget();
    static bool CloseOpenWindow(csString windowName);
    static GEMClientObject* FindItemInWorld(csString name);
	static csArray<EID> FindItemsInReach();
    static pawsSlot* GetNextFreeInvSlot();
    static bool MoveItemsToTarget(int fromContainer, int fromSlot, int toSlot, int64_t quantity);

    //training
    static bool TrainSkillMax(csString skillName);

    /**
     * Gets the ID of the currently targeted item.
     * Returns -1 if nothing is currently targeted.
     */
    static int GetTargetID();
    static EID GetTargetEID();
};



#endif /* WORLDHANDLER_H_ */
